# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import functools
import logging
import pytest
import json
import sys
import six.moves.urllib as urllib
from azure.iot.device.common.pipeline import (
    pipeline_events_base,
    pipeline_ops_base,
    pipeline_stages_base,
    pipeline_ops_mqtt,
    pipeline_events_mqtt,
)
from azure.iot.device.iothub.pipeline import (
    constant,
    pipeline_events_iothub,
    pipeline_ops_iothub,
    pipeline_stages_iothub_mqtt,
    config,
)
from azure.iot.device.iothub.pipeline.exceptions import OperationError, PipelineError
from azure.iot.device.iothub.models.message import Message
from azure.iot.device.iothub.models.methods import MethodRequest, MethodResponse
from tests.common.pipeline.helpers import (
    assert_callback_failed,
    assert_callback_succeeded,
    all_common_ops,
    all_common_events,
    all_except,
    StageTestBase,
)
from tests.iothub.pipeline.helpers import all_iothub_ops, all_iothub_events
from tests.common.pipeline import pipeline_stage_test
from azure.iot.device import constant as pkg_constant
import uuid

logging.basicConfig(level=logging.DEBUG)

this_module = sys.modules[__name__]


# This fixture makes it look like all test in this file  tests are running
# inside the pipeline thread.  Because this is an autouse fixture, we
# manually add it to the individual test.py files that need it.  If,
# instead, we had added it to some conftest.py, it would be applied to
# every tests in every file and we don't want that.
@pytest.fixture(autouse=True)
def apply_fake_pipeline_thread(fake_pipeline_thread):
    pass


fake_device_id = "__fake_device_id__"
fake_module_id = "__fake_module_id__"
fake_hostname = "__fake_hostname__"
fake_gateway_hostname = "__fake_gateway_hostname__"
fake_ca_cert = "__fake_ca_cert__"
fake_client_cert = "__fake_client_cert__"
fake_sas_token = "__fake_sas_token__"

fake_message_id = "ee9e738b-4f47-447a-9892-5b1d1d7ca5"
fake_message_id_encoded = "%24.mid=ee9e738b-4f47-447a-9892-5b1d1d7ca5"
fake_message_body = "__fake_message_body__"
fake_output_name = "__fake_output_name__"
fake_output_name_encoded = "%24.on=__fake_output_name__"
fake_content_type = "text/json"
fake_content_type_encoded = "%24.ct=text%2Fjson"
fake_content_encoding = "utf-16"
fake_content_encoding_encoded = "%24.ce=utf-16"
default_content_type = "application/json"
default_content_type_encoded = "%24.ct=application%2Fjson"
default_content_encoding_encoded = "%24.ce=utf-8"
fake_message = Message(fake_message_body)
security_message_interface_id_encoded = "%24.ifid=urn%3Aazureiot%3ASecurity%3ASecurityAgent%3A1"
fake_request_id = "__fake_request_id__"
fake_method_name = "__fake_method_name__"
fake_method_payload = "__fake_method_payload__"
fake_method_status = "__fake_method_status__"
fake_method_response = MethodResponse(
    request_id=fake_request_id, status=fake_method_status, payload=fake_method_payload
)

invalid_feature_name = "__invalid_feature_name__"
unmatched_mqtt_topic = "__unmatched_mqtt_topic__"
fake_mqtt_payload = "__fake_mqtt_payload__"

fake_c2d_topic = "devices/{}/messages/devicebound/".format(fake_device_id)
fake_c2d_topic_with_content_type = "{}{}".format(fake_c2d_topic, fake_content_type_encoded)
fake_c2d_topic_for_another_device = "devices/__other_device__/messages/devicebound/"

fake_input_name = "__fake_input_name__"
fake_input_message_topic = "devices/{}/modules/{}/inputs/{}/".format(
    fake_device_id, fake_module_id, fake_input_name
)
fake_input_message_topic_with_content_type = "{}{}".format(
    fake_input_message_topic, fake_content_type_encoded
)
fake_input_message_topic_for_another_module = "devices/{}/modules/__other_module__/messages/devicebound/".format(
    fake_device_id
)
fake_input_message_topic_for_another_device = "devices/__other_device__/modules/{}/messages/devicebound/".format(
    fake_module_id
)

fake_method_request_topic = "$iothub/methods/POST/{}/?$rid={}".format(
    fake_method_name, fake_request_id
)
fake_method_request_payload = "{}".encode("utf-8")

encoded_user_agent = urllib.parse.quote_plus(pkg_constant.USER_AGENT)

fake_message_user_property_1_key = "is-muggle"
fake_message_user_property_1_value = "yes"
fake_message_user_property_2_key = "sorted-house"
fake_message_user_property_2_value = "hufflepuff"
fake_message_user_property_1_encoded = "is-muggle=yes"
fake_message_user_property_2_encoded = "sorted-house=hufflepuff"

ops_handled_by_this_stage = [
    pipeline_ops_iothub.SetIoTHubConnectionArgsOperation,
    pipeline_ops_iothub.SendD2CMessageOperation,
    pipeline_ops_base.UpdateSasTokenOperation,
    pipeline_ops_iothub.SendOutputEventOperation,
    pipeline_ops_iothub.SendMethodResponseOperation,
    pipeline_ops_base.RequestOperation,
    pipeline_ops_base.EnableFeatureOperation,
    pipeline_ops_base.DisableFeatureOperation,
]

events_handled_by_this_stage = [pipeline_events_mqtt.IncomingMQTTMessageEvent]

pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_iothub_mqtt.IoTHubMQTTTranslationStage,
    module=this_module,
    all_ops=all_common_ops + all_iothub_ops,
    handled_ops=ops_handled_by_this_stage,
    all_events=all_common_events + all_iothub_events,
    handled_events=events_handled_by_this_stage,
    extra_initializer_defaults={"feature_to_topic": dict},
)


def create_message_with_user_properties(message_content, is_multiple):
    m = Message(message_content)
    m.custom_properties[fake_message_user_property_1_key] = fake_message_user_property_1_value
    if is_multiple:
        m.custom_properties[fake_message_user_property_2_key] = fake_message_user_property_2_value
    return m


def create_security_message(message_content):
    msg = Message(message_content)
    msg.set_as_security_message()
    return msg


def create_message_with_system_and_user_properties(message_content, is_multiple):
    if is_multiple:
        msg = Message(message_content, message_id=fake_message_id, output_name=fake_output_name)
    else:
        msg = Message(message_content, message_id=fake_message_id)

    msg.custom_properties[fake_message_user_property_1_key] = fake_message_user_property_1_value
    if is_multiple:
        msg.custom_properties[fake_message_user_property_2_key] = fake_message_user_property_2_value
    return msg


def create_security_message_with_system_and_user_properties(message_content, is_multiple):
    if is_multiple:
        msg = Message(message_content, message_id=fake_message_id, output_name=fake_output_name)
    else:
        msg = Message(message_content, message_id=fake_message_id)

    msg.custom_properties[fake_message_user_property_1_key] = fake_message_user_property_1_value
    if is_multiple:
        msg.custom_properties[fake_message_user_property_2_key] = fake_message_user_property_2_value
    msg.set_as_security_message()
    return msg


def create_message_for_output_with_user_properties(message_content, is_multiple):
    m = Message(message_content, output_name=fake_output_name)
    m.custom_properties[fake_message_user_property_1_key] = fake_message_user_property_1_value
    if is_multiple:
        m.custom_properties[fake_message_user_property_2_key] = fake_message_user_property_2_value
    return m


def create_message_for_output_with_system_and_user_properties(message_content, is_multiple):
    if is_multiple:
        msg = Message(
            message_content,
            output_name=fake_output_name,
            message_id=fake_message_id,
            content_type=fake_content_type,
        )
    else:
        msg = Message(message_content, output_name=fake_output_name, message_id=fake_message_id)

    msg.custom_properties[fake_message_user_property_1_key] = fake_message_user_property_1_value
    if is_multiple:
        msg.custom_properties[fake_message_user_property_2_key] = fake_message_user_property_2_value
    return msg


@pytest.fixture
def set_connection_args(mocker):
    return pipeline_ops_iothub.SetIoTHubConnectionArgsOperation(
        device_id=fake_device_id, hostname=fake_hostname, callback=mocker.MagicMock()
    )


@pytest.fixture
def set_connection_args_for_device(set_connection_args):
    return set_connection_args


@pytest.fixture
def set_connection_args_for_module(set_connection_args):
    set_connection_args.module_id = fake_module_id
    return set_connection_args


class IoTHubMQTTTranslationStageTestBase(StageTestBase):
    @pytest.fixture(autouse=True)
    def stage_base_configuration(self, stage, mocker):
        class NextStageForTest(pipeline_stages_base.PipelineStage):
            def _execute_op(self, op):
                pass

        next = NextStageForTest()
        root = (
            pipeline_stages_base.PipelineRootStage(config.IoTHubPipelineConfig())
            .append_stage(stage)
            .append_stage(next)
        )

        mocker.spy(stage, "_execute_op")
        mocker.spy(stage, "run_op")

        mocker.spy(next, "_execute_op")
        mocker.spy(next, "run_op")

        return root

    @pytest.fixture
    def stage(self):
        return pipeline_stages_iothub_mqtt.IoTHubMQTTTranslationStage()

    @pytest.fixture
    def stage_configured_for_device(
        self, stage, stage_base_configuration, set_connection_args_for_device, mocker
    ):
        set_connection_args_for_device.callback = None
        stage.run_op(set_connection_args_for_device)
        mocker.resetall()

    @pytest.fixture
    def stage_configured_for_module(
        self, stage, stage_base_configuration, set_connection_args_for_module, mocker
    ):
        set_connection_args_for_module.callback = None
        stage.run_op(set_connection_args_for_module)
        mocker.resetall()

    @pytest.fixture(params=["device", "module"])
    def stages_configured_for_both(
        self, request, stage, stage_base_configuration, set_connection_args, mocker
    ):
        set_connection_args.callback = None
        if request.param == "module":
            set_connection_args.module_id = fake_module_id
        stage.run_op(set_connection_args)
        mocker.resetall()


@pytest.mark.describe(
    "IoTHubMQTTTranslationStage - .run_op() -- called with SetIoTHubConnectionArgsOperation"
)
class TestIoTHubMQTTConverterWithSetAuthProviderArgs(IoTHubMQTTTranslationStageTestBase):
    @pytest.mark.it(
        "Runs a pipeline_ops_mqtt.SetMQTTConnectionArgsOperation operation on the next stage"
    )
    def test_runs_set_connection_args(self, stage, set_connection_args):
        stage.run_op(set_connection_args)
        assert stage.next._execute_op.call_count == 1
        new_op = stage.next._execute_op.call_args[0][0]
        assert isinstance(new_op, pipeline_ops_mqtt.SetMQTTConnectionArgsOperation)

    @pytest.mark.it(
        "Sets connection_args.client_id to auth_provider_args.device_id if auth_provider_args.module_id is None"
    )
    def test_sets_client_id_for_devices(self, stage, set_connection_args):
        stage.run_op(set_connection_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.client_id == fake_device_id

    @pytest.mark.it(
        "Sets connection_args.client_id to auth_provider_args.device_id/auth_provider_args.module_id if auth_provider_args.module_id is not None"
    )
    def test_sets_client_id_for_modules(self, stage, set_connection_args_for_module):
        stage.run_op(set_connection_args_for_module)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.client_id == "{}/{}".format(fake_device_id, fake_module_id)

    @pytest.mark.it(
        "Sets connection_args.hostname to auth_provider.hostname if auth_provider.gateway_hostname is None"
    )
    def test_sets_hostname_if_no_gateway(self, stage, set_connection_args):
        stage.run_op(set_connection_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.hostname == fake_hostname

    @pytest.mark.it(
        "Sets connection_args.hostname to auth_provider.gateway_hostname if auth_provider.gateway_hostname is not None"
    )
    def test_sets_hostname_if_yes_gateway(self, stage, set_connection_args):
        set_connection_args.gateway_hostname = fake_gateway_hostname
        stage.run_op(set_connection_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.hostname == fake_gateway_hostname

    @pytest.mark.it(
        "Sets connection_args.username to auth_provider.hostname/auth_provider/device_id/?api-version={api_version}&DeviceClientType={user_agent} if auth_provider_args.gateway_hostname is None and module_id is None"
    )
    def test_sets_device_username_if_no_gateway(self, stage, set_connection_args):
        stage.run_op(set_connection_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.username == "{}/{}/?api-version={}&DeviceClientType={}".format(
            fake_hostname, fake_device_id, pkg_constant.IOTHUB_API_VERSION, encoded_user_agent
        )

    @pytest.mark.it(
        "Sets connection_args.username to auth_provider.hostname/device_id/?api-version={api_version}&DeviceClientType={user_agent} if auth_provider_args.gateway_hostname is not None and module_id is None"
    )
    def test_sets_device_username_if_yes_gateway(self, stage, set_connection_args):
        set_connection_args.gateway_hostname = fake_gateway_hostname
        stage.run_op(set_connection_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.username == "{}/{}/?api-version={}&DeviceClientType={}".format(
            fake_hostname, fake_device_id, pkg_constant.IOTHUB_API_VERSION, encoded_user_agent
        )

    @pytest.mark.it(
        "Sets connection_args.username to auth_provider.hostname/auth_provider/device_id/?api-version={api_version}&DeviceClientType={user_agent} if auth_provider_args.gateway_hostname is None and module_id is None"
    )
    def test_sets_module_username_if_no_gateway(self, stage, set_connection_args_for_module):
        stage.run_op(set_connection_args_for_module)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.username == "{}/{}/{}/?api-version={}&DeviceClientType={}".format(
            fake_hostname,
            fake_device_id,
            fake_module_id,
            pkg_constant.IOTHUB_API_VERSION,
            encoded_user_agent,
        )

    @pytest.mark.it(
        "Sets connection_args.username to auth_provider.hostname/device_id/module_id/?api-version={api_version}&DeviceClientType={user_agent} if auth_provider_args.gateway_hostname is not None and module_id is None"
    )
    def test_sets_module_username_if_yes_gateway(self, stage, set_connection_args_for_module):
        set_connection_args_for_module.gateway_hostname = fake_gateway_hostname
        stage.run_op(set_connection_args_for_module)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.username == "{}/{}/{}/?api-version={}&DeviceClientType={}".format(
            fake_hostname,
            fake_device_id,
            fake_module_id,
            pkg_constant.IOTHUB_API_VERSION,
            encoded_user_agent,
        )

    @pytest.mark.it(
        "Appends product_info to connection_args.username to if self.pipeline_root.pipeline_configuration.product_info is not None"
    )
    @pytest.mark.parametrize(
        "fake_product_info, expected_product_info",
        [
            ("", ""),
            ("__fake:product:info__", "__fake%3Aproduct%3Ainfo__"),
            (4, 4),
            (
                ["fee,fi,fo,fum"],
                "%5B%27fee%2Cfi%2Cfo%2Cfum%27%5D",
            ),  # URI Encoding for str version of list
            (
                {"fake_key": "fake_value"},
                "%7B%27fake_key%27%3A%20%27fake_value%27%7D",
            ),  # URI Encoding for str version of dict
        ],
    )
    def test_appends_product_info_to_device_username(
        self, stage, set_connection_args, fake_product_info, expected_product_info
    ):
        set_connection_args.gateway_hostname = fake_gateway_hostname
        stage.pipeline_root.pipeline_configuration.product_info = fake_product_info
        stage.run_op(set_connection_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.username == "{}/{}/?api-version={}&DeviceClientType={}{}".format(
            fake_hostname,
            fake_device_id,
            pkg_constant.IOTHUB_API_VERSION,
            encoded_user_agent,
            expected_product_info,
        )

    @pytest.mark.it("Sets connection_args.ca_cert to auth_provider.ca_cert")
    def test_sets_ca_cert(self, stage, set_connection_args):
        set_connection_args.ca_cert = fake_ca_cert
        stage.run_op(set_connection_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.ca_cert == fake_ca_cert

    @pytest.mark.it("Sets connection_args.client_cert to auth_provider.client_cert")
    def test_sets_client_cert(self, stage, set_connection_args):
        set_connection_args.client_cert = fake_client_cert
        stage.run_op(set_connection_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.client_cert == fake_client_cert

    @pytest.mark.it("Sets connection_args.sas_token to auth_provider.sas_token.")
    def test_sets_sas_token(self, stage, set_connection_args):
        set_connection_args.sas_token = fake_sas_token
        stage.run_op(set_connection_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.sas_token == fake_sas_token

    @pytest.mark.it(
        "Calls the SetIoTHubConnectionArgsOperation callback with error if the pipeline_ops_mqtt.SetMQTTConnectionArgsOperation raises an Exception"
    )
    def test_set_connection_args_raises_exception(
        self, stage, mocker, arbitrary_exception, set_connection_args
    ):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_exception)
        stage.run_op(set_connection_args)
        assert_callback_failed(op=set_connection_args, error=arbitrary_exception)

    @pytest.mark.it(
        "Allows any BaseExceptions raised inside the pipeline_ops_mqtt.SetMQTTConnectionArgsOperation operation to propagate"
    )
    def test_set_connection_args_raises_base_exception(
        self, stage, mocker, arbitrary_base_exception, set_connection_args
    ):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_base_exception)
        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.run_op(set_connection_args)
        assert e_info.value is arbitrary_base_exception

    @pytest.mark.it(
        "Calls the SetIoTHubConnectionArgsOperation callback with no error if the pipeline_ops_mqtt.SetMQTTConnectionArgsOperation operation succeeds"
    )
    def test_set_connection_args_succeeds(
        self, stage, next_stage_succeeds, mocker, set_connection_args
    ):
        stage.run_op(set_connection_args)
        assert_callback_succeeded(op=set_connection_args)


@pytest.mark.describe(
    "IoTHubMQTTTranslationStage - .run_op() -- called with UpdateSasTokenOperation if the transport is disconnected"
)
class TestIoTHubMQTTConverterWithUpdateSasTokenOperationDisconnected(
    IoTHubMQTTTranslationStageTestBase
):
    @pytest.fixture
    def op(self, mocker):
        return pipeline_ops_base.UpdateSasTokenOperation(
            sas_token=fake_sas_token, callback=mocker.MagicMock()
        )

    @pytest.fixture(autouse=True)
    def transport_is_disconnected(self, stage):
        stage.pipeline_root.connected = False

    @pytest.mark.it("Immediately passes the operation to the next stage")
    def test_passes_op_immediately(self, stage, op):
        stage.run_op(op)
        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == op

    @pytest.mark.it("Completes the op with failure if some lower stage returns failure")
    def test_lower_stage_update_sas_token_fails(
        self, stage, op, next_stage_raises_arbitrary_exception, arbitrary_exception
    ):
        stage.run_op(op)
        assert_callback_failed(op=op, error=arbitrary_exception)

    @pytest.mark.it("Completes the op with success if some lower stage returns success")
    def test_lower_stage_update_sas_token_succeeds(self, stage, next_stage_succeeds, op):
        stage.run_op(op)
        assert_callback_succeeded(op=op)


@pytest.mark.describe(
    "IoTHubMQTTTranslationStage - .run_op() -- called with UpdateSasTokenOperation if the transport is connected"
)
class TestIoTHubMQTTConverterWithUpdateSasTokenOperationConnected(
    IoTHubMQTTTranslationStageTestBase
):
    @pytest.fixture
    def op(self, mocker):
        return pipeline_ops_base.UpdateSasTokenOperation(
            sas_token=fake_sas_token, callback=mocker.MagicMock()
        )

    @pytest.fixture(autouse=True)
    def transport_is_connected(self, stage):
        stage.pipeline_root.connected = True

    @pytest.mark.it("Immediately passes the operation to the next stage")
    def test_passes_op_immediately(self, stage, op):
        stage.run_op(op)
        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == op

    @pytest.mark.it(
        "Completes the op with failure if some lower stage returns failure for the UpdateSasTokenOperation"
    )
    def test_lower_stage_update_sas_token_fails(
        self, stage, op, next_stage_raises_arbitrary_exception, arbitrary_exception
    ):
        stage.run_op(op)
        assert_callback_failed(op=op, error=arbitrary_exception)

    @pytest.mark.it(
        "Passes down a ReconnectOperation instead of completing the op with success after the lower level stage returns success for the UpdateSasTokenOperation"
    )
    def test_passes_down_reconnect(self, stage, op, mocker):
        def run_op(op):
            print("in run_op {}".format(op.__class__.__name__))
            if isinstance(op, pipeline_ops_base.UpdateSasTokenOperation):
                op.callback(op, error=None)
            else:
                pass

        stage.next.run_op = mocker.MagicMock(side_effect=run_op)
        stage.run_op(op)

        assert stage.next.run_op.call_count == 2
        assert stage.next.run_op.call_args_list[0][0][0] == op
        assert isinstance(
            stage.next.run_op.call_args_list[1][0][0], pipeline_ops_base.ReconnectOperation
        )
        assert op.callback.call_count == 0

    @pytest.mark.it(
        "Completes the op with success if some lower level stage returns success for the ReconnectOperation"
    )
    def test_reconnect_succeeds(self, stage, next_stage_succeeds, op):
        # default is for stage.next.run_op to return success for all ops
        stage.run_op(op)

        assert stage.next.run_op.call_count == 2
        assert stage.next.run_op.call_args_list[0][0][0] == op
        assert isinstance(
            stage.next.run_op.call_args_list[1][0][0], pipeline_ops_base.ReconnectOperation
        )
        assert_callback_succeeded(op=op)

    @pytest.mark.it(
        "Completes the op with failure if some lower level stage returns failure for the ReconnectOperation"
    )
    def test_reconnect_fails(self, stage, op, mocker, arbitrary_exception):
        def run_op(op):
            print("in run_op {}".format(op.__class__.__name__))
            if isinstance(op, pipeline_ops_base.UpdateSasTokenOperation):
                stage.next.complete_op(op, error=None)
            elif isinstance(op, pipeline_ops_base.ReconnectOperation):
                stage.next.complete_op(op, error=arbitrary_exception)
            else:
                pass

        stage.next.run_op = mocker.MagicMock(side_effect=run_op)
        stage.run_op(op)

        assert stage.next.run_op.call_count == 2
        assert stage.next.run_op.call_args_list[0][0][0] == op
        assert isinstance(
            stage.next.run_op.call_args_list[1][0][0], pipeline_ops_base.ReconnectOperation
        )
        assert_callback_failed(op=op, error=arbitrary_exception)


basic_ops = [
    {
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {"message": fake_message, "callback": None},
        "new_op_class": pipeline_ops_mqtt.MQTTPublishOperation,
    },
    {
        "op_class": pipeline_ops_iothub.SendOutputEventOperation,
        "op_init_kwargs": {"message": fake_message, "callback": None},
        "new_op_class": pipeline_ops_mqtt.MQTTPublishOperation,
    },
    {
        "op_class": pipeline_ops_iothub.SendMethodResponseOperation,
        "op_init_kwargs": {"method_response": fake_method_response, "callback": None},
        "new_op_class": pipeline_ops_mqtt.MQTTPublishOperation,
    },
    {
        "op_class": pipeline_ops_base.EnableFeatureOperation,
        "op_init_kwargs": {"feature_name": constant.C2D_MSG, "callback": None},
        "new_op_class": pipeline_ops_mqtt.MQTTSubscribeOperation,
    },
    {
        "op_class": pipeline_ops_base.DisableFeatureOperation,
        "op_init_kwargs": {"feature_name": constant.C2D_MSG, "callback": None},
        "new_op_class": pipeline_ops_mqtt.MQTTUnsubscribeOperation,
    },
]


@pytest.mark.parametrize(
    "params",
    basic_ops,
    ids=["{}->{}".format(x["op_class"].__name__, x["new_op_class"].__name__) for x in basic_ops],
)
@pytest.mark.describe("IoTHubMQTTTranslationStage - .run_op() -- called with basic MQTT operations")
class TestIoTHubMQTTConverterBasicOperations(IoTHubMQTTTranslationStageTestBase):
    @pytest.fixture
    def op(self, params, mocker):
        op = params["op_class"](**params["op_init_kwargs"])
        op.callback = mocker.MagicMock()
        return op

    @pytest.mark.it("Runs an operation on the next stage")
    def test_runs_publish(self, params, stage, stages_configured_for_both, op):
        stage.run_op(op)
        new_op = stage.next._execute_op.call_args[0][0]
        assert isinstance(new_op, params["new_op_class"])

    @pytest.mark.it("Calls the original op callback with error if the new_op raises an exception")
    def test_operation_raises_exception(
        self, params, mocker, stage, stages_configured_for_both, op, arbitrary_exception
    ):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_exception)
        stage.run_op(op)
        assert_callback_failed(op=op, error=arbitrary_exception)

    @pytest.mark.it("Allows any any BaseExceptions raised in the new_op to propagate")
    def test_operation_raises_base_exception(
        self, params, mocker, stage, stages_configured_for_both, op, arbitrary_base_exception
    ):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_base_exception)
        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.run_op(op)
        assert e_info.value is arbitrary_base_exception

    @pytest.mark.it("Calls the original op callback with no error if the new_op operation succeeds")
    def test_operation_succeeds(
        self, params, stage, stages_configured_for_both, next_stage_succeeds, op
    ):
        stage.run_op(op)
        assert_callback_succeeded(op)


publish_ops = [
    {
        "name": "send telemetry",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {"message": Message(fake_message_body), "callback": None},
        "topic": "devices/{}/messages/events/{}&{}".format(
            fake_device_id, default_content_type_encoded, default_content_encoding_encoded
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send telemetry overriding the content type and content encoding",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {
            "message": Message(
                fake_message_body,
                content_type=fake_content_type,
                content_encoding=fake_content_encoding,
            ),
            "callback": None,
        },
        "topic": "devices/{}/messages/events/{}&{}".format(
            fake_device_id, fake_content_type_encoded, fake_content_encoding_encoded
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send telemetry overriding only the content type",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {
            "message": Message(fake_message_body, content_type=fake_content_type),
            "callback": None,
        },
        "topic": "devices/{}/messages/events/{}&{}".format(
            fake_device_id, fake_content_type_encoded, default_content_encoding_encoded
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send telemetry with single system property",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {
            "message": Message(fake_message_body, output_name=fake_output_name),
            "callback": None,
        },
        "topic": "devices/{}/messages/events/{}&{}&{}".format(
            fake_device_id,
            fake_output_name_encoded,
            default_content_type_encoded,
            default_content_encoding_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send security message",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {"message": create_security_message(fake_message_body), "callback": None},
        "topic": "devices/{}/messages/events/{}&{}&{}".format(
            fake_device_id,
            default_content_type_encoded,
            default_content_encoding_encoded,
            security_message_interface_id_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send telemetry with multiple system properties",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {
            "message": Message(
                fake_message_body, message_id=fake_message_id, output_name=fake_output_name
            ),
            "callback": None,
        },
        "topic": "devices/{}/messages/events/{}&{}&{}&{}".format(
            fake_device_id,
            fake_output_name_encoded,
            fake_message_id_encoded,
            default_content_type_encoded,
            default_content_encoding_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send telemetry with only single user property",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {
            "message": create_message_with_user_properties(fake_message_body, is_multiple=False),
            "callback": None,
        },
        "topic": "devices/{}/messages/events/{}&{}&{}".format(
            fake_device_id,
            default_content_type_encoded,
            default_content_encoding_encoded,
            fake_message_user_property_1_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send telemetry with only multiple user properties",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {
            "message": create_message_with_user_properties(fake_message_body, is_multiple=True),
            "callback": None,
        },
        # For more than 1 user property the order could be different, creating 2 different topics
        "topic1": "devices/{}/messages/events/{}&{}&{}&{}".format(
            fake_device_id,
            default_content_type_encoded,
            default_content_encoding_encoded,
            fake_message_user_property_1_encoded,
            fake_message_user_property_2_encoded,
        ),
        "topic2": "devices/{}/messages/events/{}&{}&{}&{}".format(
            fake_device_id,
            default_content_type_encoded,
            default_content_encoding_encoded,
            fake_message_user_property_2_encoded,
            fake_message_user_property_1_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send telemetry with 1 system and 1 user property",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {
            "message": create_message_with_system_and_user_properties(
                fake_message_body, is_multiple=False
            ),
            "callback": None,
        },
        "topic": "devices/{}/messages/events/{}&{}&{}&{}".format(
            fake_device_id,
            fake_message_id_encoded,
            default_content_type_encoded,
            default_content_encoding_encoded,
            fake_message_user_property_1_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send telemetry with multiple system and multiple user properties",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {
            "message": create_message_with_system_and_user_properties(
                fake_message_body, is_multiple=True
            ),
            "callback": None,
        },
        # For more than 1 user property the order could be different, creating 2 different topics
        "topic1": "devices/{}/messages/events/{}&{}&{}&{}&{}&{}".format(
            fake_device_id,
            fake_output_name_encoded,
            fake_message_id_encoded,
            default_content_type_encoded,
            default_content_encoding_encoded,
            fake_message_user_property_1_encoded,
            fake_message_user_property_2_encoded,
        ),
        "topic2": "devices/{}/messages/events/{}&{}&{}&{}&{}&{}".format(
            fake_device_id,
            fake_output_name_encoded,
            fake_message_id_encoded,
            default_content_type_encoded,
            default_content_encoding_encoded,
            fake_message_user_property_2_encoded,
            fake_message_user_property_1_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send security message with multiple system and multiple user properties",
        "stage_type": "device",
        "op_class": pipeline_ops_iothub.SendD2CMessageOperation,
        "op_init_kwargs": {
            "message": create_security_message_with_system_and_user_properties(
                fake_message_body, is_multiple=True
            ),
            "callback": None,
        },
        # For more than 1 user property the order could be different, creating 2 different topics
        "topic1": "devices/{}/messages/events/{}&{}&{}&{}&{}&{}&{}".format(
            fake_device_id,
            fake_output_name_encoded,
            fake_message_id_encoded,
            default_content_type_encoded,
            default_content_encoding_encoded,
            security_message_interface_id_encoded,
            fake_message_user_property_1_encoded,
            fake_message_user_property_2_encoded,
        ),
        "topic2": "devices/{}/messages/events/{}&{}&{}&{}&{}&{}&{}".format(
            fake_device_id,
            fake_output_name_encoded,
            fake_message_id_encoded,
            default_content_type_encoded,
            default_content_encoding_encoded,
            security_message_interface_id_encoded,
            fake_message_user_property_2_encoded,
            fake_message_user_property_1_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send output",
        "stage_type": "module",
        "op_class": pipeline_ops_iothub.SendOutputEventOperation,
        "op_init_kwargs": {
            "message": Message(fake_message_body, output_name=fake_output_name),
            "callback": None,
        },
        "topic": "devices/{}/modules/{}/messages/events/%24.on={}&{}&{}".format(
            fake_device_id,
            fake_module_id,
            fake_output_name,
            default_content_type_encoded,
            default_content_encoding_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send output overriding content type and content encoding",
        "stage_type": "module",
        "op_class": pipeline_ops_iothub.SendOutputEventOperation,
        "op_init_kwargs": {
            "message": Message(
                fake_message_body,
                output_name=fake_output_name,
                content_type=fake_content_type,
                content_encoding=fake_content_encoding,
            ),
            "callback": None,
        },
        "topic": "devices/{}/modules/{}/messages/events/%24.on={}&{}&{}".format(
            fake_device_id,
            fake_module_id,
            fake_output_name,
            fake_content_type_encoded,
            fake_content_encoding_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send output with system properties",
        "stage_type": "module",
        "op_class": pipeline_ops_iothub.SendOutputEventOperation,
        "op_init_kwargs": {
            "message": Message(
                fake_message_body, message_id=fake_message_id, output_name=fake_output_name
            ),
            "callback": None,
        },
        "topic": "devices/{}/modules/{}/messages/events/%24.on={}&{}&{}&{}".format(
            fake_device_id,
            fake_module_id,
            fake_output_name,
            fake_message_id_encoded,
            default_content_type_encoded,
            default_content_encoding_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send output with only 1 user property",
        "stage_type": "module",
        "op_class": pipeline_ops_iothub.SendOutputEventOperation,
        "op_init_kwargs": {
            "message": create_message_for_output_with_user_properties(
                fake_message_body, is_multiple=False
            ),
            "callback": None,
        },
        "topic": "devices/{}/modules/{}/messages/events/%24.on={}&{}&{}&{}".format(
            fake_device_id,
            fake_module_id,
            fake_output_name,
            default_content_type_encoded,
            default_content_encoding_encoded,
            fake_message_user_property_1_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send output with only multiple user properties",
        "stage_type": "module",
        "op_class": pipeline_ops_iothub.SendOutputEventOperation,
        "op_init_kwargs": {
            "message": create_message_for_output_with_user_properties(
                fake_message_body, is_multiple=True
            ),
            "callback": None,
        },
        "topic1": "devices/{}/modules/{}/messages/events/%24.on={}&{}&{}&{}&{}".format(
            fake_device_id,
            fake_module_id,
            fake_output_name,
            default_content_type_encoded,
            default_content_encoding_encoded,
            fake_message_user_property_1_encoded,
            fake_message_user_property_2_encoded,
        ),
        "topic2": "devices/{}/modules/{}/messages/events/%24.on={}&{}&{}&{}&{}".format(
            fake_device_id,
            fake_module_id,
            fake_output_name,
            default_content_type_encoded,
            default_content_encoding_encoded,
            fake_message_user_property_2_encoded,
            fake_message_user_property_1_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send output with 1 system and 1 user property",
        "stage_type": "module",
        "op_class": pipeline_ops_iothub.SendOutputEventOperation,
        "op_init_kwargs": {
            "message": create_message_for_output_with_system_and_user_properties(
                fake_message_body, is_multiple=False
            ),
            "callback": None,
        },
        "topic": "devices/{}/modules/{}/messages/events/%24.on={}&{}&{}&{}&{}".format(
            fake_device_id,
            fake_module_id,
            fake_output_name,
            fake_message_id_encoded,
            default_content_type_encoded,
            default_content_encoding_encoded,
            fake_message_user_property_1_encoded,
        ),
        "publish_payload": fake_message_body,
    },
    {
        "name": "send method result",
        "stage_type": "both",
        "op_class": pipeline_ops_iothub.SendMethodResponseOperation,
        "op_init_kwargs": {"method_response": fake_method_response, "callback": None},
        "topic": "$iothub/methods/res/__fake_method_status__/?$rid=__fake_request_id__",
        "publish_payload": json.dumps(fake_method_payload),
    },
]


@pytest.mark.parametrize("params", publish_ops, ids=[x["name"] for x in publish_ops])
@pytest.mark.describe("IoTHubMQTTTranslationStage - .run_op() -- called with publish operations")
class TestIoTHubMQTTConverterForPublishOps(IoTHubMQTTTranslationStageTestBase):
    @pytest.fixture
    def op(self, params, mocker):
        op = params["op_class"](**params["op_init_kwargs"])
        op.callback = mocker.MagicMock()
        return op

    @pytest.mark.it("Uses the correct topic and encodes message properties string when publishing")
    def test_uses_device_topic_for_devices(self, stage, stages_configured_for_both, params, op):
        if params["stage_type"] == "device" and stage.module_id:
            pytest.skip()
        elif params["stage_type"] == "module" and not stage.module_id:
            pytest.skip()
        stage.run_op(op)
        new_op = stage.next._execute_op.call_args[0][0]
        if "multiple user properties" in params["name"]:
            assert new_op.topic == params["topic1"] or new_op.topic == params["topic2"]
        else:
            assert new_op.topic == params["topic"]

    @pytest.mark.it("Sends the body in the payload of the MQTT publish operation")
    def test_sends_correct_body(self, stage, stages_configured_for_both, params, op):
        stage.run_op(op)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.payload == params["publish_payload"]


feature_name_to_subscribe_topic = [
    {
        "stage_type": "device",
        "feature_name": constant.C2D_MSG,
        "topic": "devices/{}/messages/devicebound/#".format(fake_device_id),
    },
    {
        "stage_type": "module",
        "feature_name": constant.INPUT_MSG,
        "topic": "devices/{}/modules/{}/inputs/#".format(fake_device_id, fake_module_id),
    },
    {"stage_type": "both", "feature_name": constant.METHODS, "topic": "$iothub/methods/POST/#"},
]

sub_unsub_operations = [
    {
        "op_class": pipeline_ops_base.EnableFeatureOperation,
        "new_op": pipeline_ops_mqtt.MQTTSubscribeOperation,
    },
    {
        "op_class": pipeline_ops_base.DisableFeatureOperation,
        "new_op": pipeline_ops_mqtt.MQTTUnsubscribeOperation,
    },
]


@pytest.mark.describe(
    "IoTHubMQTTTranslationStage - .run_op() -- called with EnableFeature or DisableFeature"
)
class TestIoTHubMQTTConverterWithEnableFeature(IoTHubMQTTTranslationStageTestBase):
    @pytest.mark.parametrize(
        "topic_parameters",
        feature_name_to_subscribe_topic,
        ids=[
            "{} {}".format(x["stage_type"], x["feature_name"])
            for x in feature_name_to_subscribe_topic
        ],
    )
    @pytest.mark.parametrize(
        "op_parameters",
        sub_unsub_operations,
        ids=[x["op_class"].__name__ for x in sub_unsub_operations],
    )
    @pytest.mark.it("Converts the feature_name to the correct topic")
    def test_converts_feature_name_to_topic(
        self, mocker, stage, stages_configured_for_both, topic_parameters, op_parameters
    ):
        if topic_parameters["stage_type"] == "device" and stage.module_id:
            pytest.skip()
        elif topic_parameters["stage_type"] == "module" and not stage.module_id:
            pytest.skip()
        stage.next._execute_op = mocker.Mock()
        op = op_parameters["op_class"](
            feature_name=topic_parameters["feature_name"], callback=mocker.MagicMock()
        )
        stage.run_op(op)
        new_op = stage.next._execute_op.call_args[0][0]
        assert isinstance(new_op, op_parameters["new_op"])
        assert new_op.topic == topic_parameters["topic"]

    @pytest.mark.it("Fails on an invalid feature_name")
    @pytest.mark.parametrize(
        "op_parameters",
        sub_unsub_operations,
        ids=[x["op_class"].__name__ for x in sub_unsub_operations],
    )
    def test_fails_on_invalid_feature_name(
        self, mocker, stage, stages_configured_for_both, op_parameters
    ):
        op = op_parameters["op_class"](
            feature_name=invalid_feature_name, callback=mocker.MagicMock()
        )
        stage.run_op(op)
        assert_callback_failed(op=op, error=KeyError)


@pytest.fixture
def add_pipeline_root(stage, mocker):
    root = pipeline_stages_base.PipelineRootStage(mocker.MagicMock())
    mocker.spy(root, "handle_pipeline_event")
    stage.previous = root
    stage.pipeline_root = root


@pytest.mark.describe(
    "IoTHubMQTTTranslationStage - .handle_pipeline_event() -- called with unmatched topic"
)
class TestIoTHubMQTTConverterHandlePipelineEvent(IoTHubMQTTTranslationStageTestBase):
    @pytest.mark.it("Passes up any mqtt messages with topics that aren't matched by this stage")
    def test_passes_up_mqtt_message_with_unknown_topic(
        self, stage, stages_configured_for_both, add_pipeline_root, mocker
    ):
        event = pipeline_events_mqtt.IncomingMQTTMessageEvent(
            topic=unmatched_mqtt_topic, payload=fake_mqtt_payload
        )
        stage.handle_pipeline_event(event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        assert stage.previous.handle_pipeline_event.call_args == mocker.call(event)


@pytest.fixture
def c2d_event():
    return pipeline_events_mqtt.IncomingMQTTMessageEvent(
        topic=fake_c2d_topic, payload=fake_mqtt_payload
    )


@pytest.mark.describe(
    "IoTHubMQTTTranslationStage - .handle_pipeline_event() -- called with C2D topic"
)
class TestIoTHubMQTTConverterHandlePipelineEventC2D(IoTHubMQTTTranslationStageTestBase):
    @pytest.mark.it(
        "Converts mqtt message with topic devices/device_id/message/devicebound/ to c2d event"
    )
    def test_converts_c2d_topic_to_c2d_events(
        self, mocker, stage, stage_configured_for_device, add_pipeline_root, c2d_event
    ):
        stage.handle_pipeline_event(c2d_event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert isinstance(new_event, pipeline_events_iothub.C2DMessageEvent)

    @pytest.mark.it("Convers the mqtt payload of a c2d message into a Message object")
    def test_creates_message_object_for_c2d_event(
        self, mocker, stage, stage_configured_for_device, add_pipeline_root, c2d_event
    ):
        stage.handle_pipeline_event(c2d_event)
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert isinstance(new_event.message, Message)

    @pytest.mark.it("Extracts message properties from the mqtt topic for c2d messages")
    def test_extracts_c2d_message_properties_from_topic_name(
        self, mocker, stage, stage_configured_for_device, add_pipeline_root
    ):
        event = pipeline_events_mqtt.IncomingMQTTMessageEvent(
            topic=fake_c2d_topic_with_content_type, payload=fake_mqtt_payload
        )
        stage.handle_pipeline_event(event)
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert new_event.message.content_type == fake_content_type

    @pytest.mark.it("Passes up c2d messages destined for another device")
    def test_if_topic_is_c2d_for_another_device(
        self, mocker, stage, stage_configured_for_device, add_pipeline_root
    ):
        event = pipeline_events_mqtt.IncomingMQTTMessageEvent(
            topic=fake_c2d_topic_for_another_device, payload=fake_mqtt_payload
        )
        stage.handle_pipeline_event(event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        assert stage.previous.handle_pipeline_event.call_args == mocker.call(event)


@pytest.mark.describe("IotHubMQTTConverter - .run_op() -- called with RequestOperation")
class TestIotHubMQTTConverterWithSendIotRequest(IoTHubMQTTTranslationStageTestBase):
    @pytest.fixture
    def fake_request_type(self):
        return "twin"

    @pytest.fixture
    def fake_method(self):
        return "__fake_method__"

    @pytest.fixture
    def fake_resource_location(self):
        return "__fake_resource_location__"

    @pytest.fixture
    def fake_request_body(self):
        return "__fake_request_body__"

    @pytest.fixture
    def fake_request_body_as_string(self, fake_request_body):
        return json.dumps(fake_request_body)

    @pytest.fixture
    def fake_request_id(self):
        return "__fake_request_id__"

    @pytest.fixture
    def op(
        self,
        fake_request_type,
        fake_method,
        fake_resource_location,
        fake_request_body,
        fake_request_id,
        mocker,
    ):
        return pipeline_ops_base.RequestOperation(
            request_type=fake_request_type,
            method=fake_method,
            resource_location=fake_resource_location,
            request_body=fake_request_body,
            request_id=fake_request_id,
            callback=mocker.MagicMock(),
        )

    @pytest.mark.it("calls the op callback with an OperationError if request_type is not 'twin'")
    def test_sends_bad_request_type(self, stage, op):
        op.request_type = "not_twin"
        stage.run_op(op)
        assert_callback_failed(op=op, error=OperationError)

    @pytest.mark.it(
        "Runs an MQTTPublishOperation on the next stage with the topic formated as '$iothub/twin/{method}{resource_location}?$rid={request_id}' and the payload as the request_body"
    )
    def test_sends_new_operation(
        self, stage, op, fake_method, fake_resource_location, fake_request_id, fake_request_body
    ):
        stage.run_op(op)
        assert stage.next.run_op.call_count == 1
        new_op = stage.next.run_op.call_args[0][0]
        assert isinstance(new_op, pipeline_ops_mqtt.MQTTPublishOperation)
        assert new_op.topic == "$iothub/twin/{method}{resource_location}?$rid={request_id}".format(
            method=fake_method, resource_location=fake_resource_location, request_id=fake_request_id
        )
        assert new_op.payload == fake_request_body

    @pytest.mark.it("Returns an OperationError through the op callback if there is no next stage")
    def test_runs_with_no_next_stage(self, stage, op):
        stage.next = None
        stage.run_op(op)
        assert_callback_failed(op=op, error=PipelineError)

    @pytest.mark.it(
        "Handles any Exceptions raised by the MQTTPublishOperation and returns them through the op callback"
    )
    def test_next_stage_raises_exception(self, mocker, stage, op, arbitrary_exception):
        # Although stage.next.run_op is already a mocker.spy (i.e. a MagicMock) as a result of the
        # fixture config, in Python 3.4 setting the side effect directly results in a TypeError
        # (it is unclear as to why at this time)
        stage.next.run_op = mocker.MagicMock(side_effect=arbitrary_exception)
        stage.run_op(op)
        assert_callback_failed(op=op, error=arbitrary_exception)

    @pytest.mark.it("Allows any BaseExceptions raised by the MQTTPublishOperation to propagate")
    def test_next_stage_raises_base_exception(self, mocker, stage, op, arbitrary_base_exception):
        # Although stage.next.run_op is already a mocker.spy (i.e. a MagicMock) as a result of the
        # fixture config, in Python 3.4 setting the side effect directly results in a TypeError
        # (it is unclear as to why at this time)
        stage.next.run_op = mocker.MagicMock(side_effect=arbitrary_base_exception)
        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.run_op(op)
        assert e_info.value is arbitrary_base_exception

    @pytest.mark.it(
        "Returns error as the MQTTPublishOperation error in the op callback if the MQTTPublishOperation returned an error in its operation callback"
    )
    def test_publish_op_returns_failure(self, stage, op, arbitrary_exception):
        def next_stage_run_op(self, op):
            op.callback(op, error=arbitrary_exception)

        stage.next.run_op = functools.partial(next_stage_run_op, (stage.next,))
        stage.run_op(op)
        assert_callback_failed(op=op, error=arbitrary_exception)

    @pytest.mark.it(
        "Returns error=None in the operation callback if the MQTTPublishOperation returned error=None in its operation callback"
    )
    def test_publish_op_returns_success(self, stage, op):
        def next_stage_run_op(self, op):
            op.callback(op, error=None)

        stage.next.run_op = functools.partial(next_stage_run_op, (stage.next,))
        stage.run_op(op)
        assert_callback_succeeded(op=op)


@pytest.fixture
def input_message_event():
    return pipeline_events_mqtt.IncomingMQTTMessageEvent(
        topic=fake_input_message_topic, payload=fake_mqtt_payload
    )


@pytest.mark.describe(
    "IoTHubMQTTTranslationStage - .handle_pipeline_event() -- called with input message topic"
)
class TestIoTHubMQTTConverterHandlePipelineEventInputMessages(IoTHubMQTTTranslationStageTestBase):
    @pytest.mark.it(
        "Converts mqtt message with topic devices/device_id/modules/module_id/inputs/input_name/ to input event"
    )
    def test_converts_input_topic_to_input_event(
        self, mocker, stage, stage_configured_for_module, add_pipeline_root, input_message_event
    ):
        stage.handle_pipeline_event(input_message_event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert isinstance(new_event, pipeline_events_iothub.InputMessageEvent)

    @pytest.mark.it("Converts the mqtt payload of an input message into a Message object")
    def test_creates_message_object_for_input_event(
        self, mocker, stage, stage_configured_for_module, add_pipeline_root, input_message_event
    ):
        stage.handle_pipeline_event(input_message_event)
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert isinstance(new_event.message, Message)

    @pytest.mark.it("Extracts the input name of an input message from the mqtt topic")
    def test_extracts_input_name_from_topic(
        self, mocker, stage, stage_configured_for_module, add_pipeline_root, input_message_event
    ):
        stage.handle_pipeline_event(input_message_event)
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert new_event.input_name == fake_input_name

    @pytest.mark.it("Extracts message properties from the mqtt topic for input messages")
    def test_extracts_input_message_properties_from_topic_name(
        self, mocker, stage, stage_configured_for_module, add_pipeline_root
    ):
        event = pipeline_events_mqtt.IncomingMQTTMessageEvent(
            topic=fake_input_message_topic_with_content_type, payload=fake_mqtt_payload
        )
        stage.handle_pipeline_event(event)
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert new_event.message.content_type == fake_content_type

    @pytest.mark.parametrize(
        "topic",
        [fake_input_message_topic_for_another_device, fake_input_message_topic_for_another_module],
        ids=["different device_id", "same device_id"],
    )
    @pytest.mark.it("Passes up input messages destined for another module")
    def test_if_topic_is_input_message_for_another_module(
        self, mocker, stage, stage_configured_for_module, add_pipeline_root, topic
    ):
        event = pipeline_events_mqtt.IncomingMQTTMessageEvent(
            topic=topic, payload=fake_mqtt_payload
        )
        stage.handle_pipeline_event(event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        assert stage.previous.handle_pipeline_event.call_args == mocker.call(event)


@pytest.fixture
def method_request_event():
    return pipeline_events_mqtt.IncomingMQTTMessageEvent(
        topic=fake_method_request_topic, payload=fake_method_request_payload
    )


@pytest.mark.describe(
    "IoTHubMQTTTranslationStage - .handle_pipeline_event() -- called with method request topic"
)
class TestIoTHubMQTTConverterHandlePipelineEventMethodRequets(IoTHubMQTTTranslationStageTestBase):
    @pytest.mark.it(
        "Converts mqtt messages with topic $iothub/methods/POST/{method name}/?$rid={request id} to method request events"
    )
    def test_converts_method_request_topic_to_method_request_event(
        self, mocker, stage, stages_configured_for_both, add_pipeline_root, method_request_event
    ):
        stage.handle_pipeline_event(method_request_event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert isinstance(new_event, pipeline_events_iothub.MethodRequestEvent)

    @pytest.mark.it("Makes a MethodRequest object to hold the method request details")
    def test_passes_method_request_object_in_method_request_event(
        self, mocker, stage, stages_configured_for_both, add_pipeline_root, method_request_event
    ):
        stage.handle_pipeline_event(method_request_event)
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert isinstance(new_event.method_request, MethodRequest)

    @pytest.mark.it("Extracts the method name from the mqtt topic")
    def test_extracts_method_name_from_method_request_topic(
        self, mocker, stage, stages_configured_for_both, add_pipeline_root, method_request_event
    ):
        stage.handle_pipeline_event(method_request_event)
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert new_event.method_request.name == fake_method_name

    @pytest.mark.it("Extracts the request id from the mqtt topic")
    def test_extracts_request_id_from_method_request_topic(
        self, mocker, stage, stages_configured_for_both, add_pipeline_root, method_request_event
    ):
        stage.handle_pipeline_event(method_request_event)
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert new_event.method_request.request_id == fake_request_id

    @pytest.mark.it(
        "Puts the payload of the mqtt message as the payload of the method requets object"
    )
    def test_puts_mqtt_payload_in_method_request_payload(
        self, mocker, stage, stages_configured_for_both, add_pipeline_root, method_request_event
    ):
        stage.handle_pipeline_event(method_request_event)
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert new_event.method_request.payload == json.loads(
            fake_method_request_payload.decode("utf-8")
        )


@pytest.mark.describe(
    "IotHubMQTTConverter - .handle_pipeline_event() -- called with twin response topic"
)
class TestIotHubMQTTConverterHandlePipelineEventTwinResponse(IoTHubMQTTTranslationStageTestBase):
    @pytest.fixture
    def fake_request_id(self):
        return "__fake_request_id__"

    @pytest.fixture
    def fake_status_code(self):
        return 200

    @pytest.fixture
    def bad_status_code(self):
        return "__bad_status_code__"

    @pytest.fixture
    def fake_topic_name(self, fake_request_id, fake_status_code):
        return "$iothub/twin/res/{status_code}/?$rid={request_id}".format(
            status_code=fake_status_code, request_id=fake_request_id
        )

    @pytest.fixture
    def fake_topic_name_with_missing_request_id(self, fake_status_code):
        return "$iothub/twin/res/{status_code}".format(status_code=fake_status_code)

    @pytest.fixture
    def fake_topic_name_with_missing_status_code(self, fake_request_id):
        return "$iothub/twin/res/?$rid={request_id}".format(request_id=fake_request_id)

    @pytest.fixture
    def fake_topic_name_with_bad_status_code(self, fake_request_id, bad_status_code):
        return "$iothub/twin/res/{status_code}/?$rid={request_id}".format(
            request_id=fake_request_id, status_code=bad_status_code
        )

    @pytest.fixture
    def fake_payload(self):
        return "__fake_payload__"

    @pytest.fixture
    def fake_event(self, fake_topic_name, fake_payload):
        return pipeline_events_mqtt.IncomingMQTTMessageEvent(
            topic=fake_topic_name, payload=fake_payload
        )

    @pytest.fixture
    def fixup_stage_for_test(self, stage, add_pipeline_root):
        print("Adding module")
        stage.module_id = fake_module_id
        stage.device_id = fake_device_id

    @pytest.mark.it(
        "Calls .handle_pipeline_event() on the previous stage with an ResponseEvent, with request_id and status_code as attributes extracted from the topic and the response_body attirbute set to the payload"
    )
    def test_extracts_request_id_status_code_and_payload(
        self,
        stage,
        fixup_stage_for_test,
        fake_request_id,
        fake_status_code,
        fake_payload,
        fake_event,
    ):
        stage.handle_pipeline_event(event=fake_event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert isinstance(new_event, pipeline_events_base.ResponseEvent)
        assert new_event.status_code == fake_status_code
        assert new_event.request_id == fake_request_id
        assert new_event.response_body == fake_payload

    @pytest.mark.it(
        "Calls the unhandled exception handler with a PipelineError if there is no previous stage"
    )
    def test_no_previous_stage(
        self, stage, fixup_stage_for_test, fake_event, unhandled_error_handler
    ):
        stage.previous = None
        stage.handle_pipeline_event(fake_event)
        assert unhandled_error_handler.call_count == 1
        assert isinstance(unhandled_error_handler.call_args[0][0], PipelineError)

    @pytest.mark.it(
        "Calls the unhandled exception handler if the requet_id is missing from the topic name"
    )
    def test_invalid_topic_with_missing_request_id(
        self,
        stage,
        fixup_stage_for_test,
        fake_event,
        fake_topic_name_with_missing_request_id,
        unhandled_error_handler,
    ):
        fake_event.topic = fake_topic_name_with_missing_request_id
        stage.handle_pipeline_event(event=fake_event)
        assert unhandled_error_handler.call_count == 1
        assert isinstance(unhandled_error_handler.call_args[0][0], IndexError)

    @pytest.mark.it(
        "Calls the unhandled exception handler if the status code is missing from the topic name"
    )
    def test_invlid_topic_with_missing_status_code(
        self,
        stage,
        fixup_stage_for_test,
        fake_event,
        fake_topic_name_with_missing_status_code,
        unhandled_error_handler,
    ):
        fake_event.topic = fake_topic_name_with_missing_status_code
        stage.handle_pipeline_event(event=fake_event)
        assert unhandled_error_handler.call_count == 1
        assert isinstance(unhandled_error_handler.call_args[0][0], ValueError)

    @pytest.mark.it(
        "Calls the unhandled exception handler if the status code in the topic name is not numeric"
    )
    def test_invlid_topic_with_bad_status_code(
        self,
        stage,
        fixup_stage_for_test,
        fake_event,
        fake_topic_name_with_bad_status_code,
        unhandled_error_handler,
    ):
        fake_event.topic = fake_topic_name_with_bad_status_code
        stage.handle_pipeline_event(event=fake_event)
        assert unhandled_error_handler.call_count == 1
        assert isinstance(unhandled_error_handler.call_args[0][0], ValueError)


@pytest.mark.describe(
    "IotHubMQTTConverter - .handle_pipeline_event() -- called with twin patch topic"
)
class TestIotHubMQTTConverterHandlePipelineEventTwinPatch(IoTHubMQTTTranslationStageTestBase):
    @pytest.fixture
    def fake_topic_name(self):
        return "$iothub/twin/PATCH/properties/desired"

    @pytest.fixture
    def fake_patch(self):
        return {"__fake_patch__": "yes"}

    @pytest.fixture
    def fake_patch_as_bytes(self, fake_patch):
        return json.dumps(fake_patch).encode("utf-8")

    @pytest.fixture
    def fake_patch_not_bytes(self):
        return "__fake_patch_that_is_not_bytes__"

    @pytest.fixture
    def fake_patch_not_json(self):
        return "__fake_patch_that_is_not_json__".encode("utf-8")

    @pytest.fixture
    def fake_event(self, fake_topic_name, fake_patch_as_bytes):
        return pipeline_events_mqtt.IncomingMQTTMessageEvent(
            topic=fake_topic_name, payload=fake_patch_as_bytes
        )

    @pytest.fixture
    def fixup_stage_for_test(self, stage, add_pipeline_root):
        print("Adding module")
        stage.module_id = fake_module_id
        stage.device_id = fake_device_id

    @pytest.mark.it(
        "Calls .handle_pipeline_event() on the previous stage with an TwinDesiredPropertiesPatchEvent, with the patch set to the payload after decoding and deserializing it"
    )
    def test_calls_previous_stage(self, stage, fixup_stage_for_test, fake_event, fake_patch):
        stage.handle_pipeline_event(fake_event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert isinstance(new_event, pipeline_events_iothub.TwinDesiredPropertiesPatchEvent)
        assert new_event.patch == fake_patch

    @pytest.mark.it(
        "Calls the unhandled exception handler with a PipelineError if there is no previous stage"
    )
    def test_no_previous_stage(
        self, stage, fixup_stage_for_test, fake_event, unhandled_error_handler
    ):
        stage.previous = None
        stage.handle_pipeline_event(fake_event)
        assert unhandled_error_handler.call_count == 1
        assert isinstance(unhandled_error_handler.call_args[0][0], PipelineError)

    @pytest.mark.it("Calls the unhandled exception handler if the payload is not a Bytes object")
    def test_payload_not_bytes(
        self, stage, fixup_stage_for_test, fake_event, fake_patch_not_bytes, unhandled_error_handler
    ):
        fake_event.payload = fake_patch_not_bytes
        stage.handle_pipeline_event(fake_event)
        assert unhandled_error_handler.call_count == 1
        if not (
            isinstance(unhandled_error_handler.call_args[0][0], AttributeError)
            or isinstance(unhandled_error_handler.call_args[0][0], ValueError)
        ):
            assert False

    @pytest.mark.it(
        "Calls the unhandled exception handler if the payload cannot be deserialized as a JSON object"
    )
    def test_payload_not_json(
        self, stage, fixup_stage_for_test, fake_event, fake_patch_not_json, unhandled_error_handler
    ):
        fake_event.payload = fake_patch_not_json
        stage.handle_pipeline_event(fake_event)
        assert unhandled_error_handler.call_count == 1
        assert isinstance(unhandled_error_handler.call_args[0][0], ValueError)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import six.moves.urllib as urllib

logger = logging.getLogger(__name__)


def _get_topic_base():
    """
    return the string that creates the beginning of all topics for DPS
    """
    return "$dps/registrations/"


def get_topic_for_subscribe():
    """
    return the topic string used to subscribe for receiving future responses from DPS
    """
    return _get_topic_base() + "res/#"


def get_topic_for_register(request_id):
    """
    return the topic string used to publish telemetry
    """
    return (_get_topic_base() + "PUT/iotdps-register/?$rid={request_id}").format(
        request_id=request_id
    )


def get_topic_for_query(request_id, operation_id):
    """
    :return: The topic for cloud to device messages.It is of the format
    "devices/<deviceid>/messages/devicebound/#"
    """
    return (
        _get_topic_base()
        + "GET/iotdps-get-operationstatus/?$rid={request_id}&operationId={operation_id}"
    ).format(request_id=request_id, operation_id=operation_id)


def get_topic_for_response():
    """
    return the topic string used to publish telemetry
    """
    return _get_topic_base() + "res/"


def is_query_topic(topic):
    if "GET/iotdps-get-operationstatus" in topic:
        return True
    return False


def is_dps_response_topic(topic):
    """
    Topics for responses from DPS are of the following format:
    $dps/registrations/res/<statuscode>/?$<key1>=<value1>&<key2>=<value2>&<key3>=<value3>
    :param topic: The topic string
    """
    if get_topic_for_response() in topic:
        return True
    return False


def extract_properties_from_topic(topic):
    """
    Topics for responses from DPS are of the following format:
    $dps/registrations/res/<statuscode>/?$<key1>=<value1>&<key2>=<value2>&<key3>=<value3>
    Extract key=value pairs from the latter part of the topic.
    :param topic: The topic string
    :return key_values_dict : a dictionary of key mapped to a list of values.
    """
    topic_parts = topic.split("$")
    key_value_dict = urllib.parse.parse_qs(topic_parts[2])
    return key_value_dict


def extract_status_code_from_topic(topic):
    """
    Topics for responses from DPS are of the following format:
    $dps/registrations/res/<statuscode>/?$<key1>=<value1>&<key2>=<value2>&<key3>=<value3>
    Extract the status code part from the topic.
    :param topic: The topic string
    """
    POS_STATUS_CODE_IN_TOPIC = 3
    topic_parts = topic.split("$")
    url_parts = topic_parts[1].split("/")
    status_code = url_parts[POS_STATUS_CODE_IN_TOPIC]
    return status_code



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import sys
import logging
from azure.iot.device.common.pipeline import pipeline_ops_mqtt
from tests.common.pipeline import pipeline_data_object_test

logging.basicConfig(level=logging.DEBUG)
this_module = sys.modules[__name__]

pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_mqtt.SetMQTTConnectionArgsOperation,
    module=this_module,
    positional_arguments=["client_id", "hostname", "username", "callback"],
    keyword_arguments={"ca_cert": None, "client_cert": None, "sas_token": None},
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_mqtt.MQTTPublishOperation,
    module=this_module,
    positional_arguments=["topic", "payload", "callback"],
    extra_defaults={"needs_connection": True},
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_mqtt.MQTTSubscribeOperation,
    module=this_module,
    positional_arguments=["topic", "callback"],
    extra_defaults={"needs_connection": True, "retry_timer": None, "timeout_timer": None},
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_mqtt.MQTTUnsubscribeOperation,
    module=this_module,
    positional_arguments=["topic", "callback"],
    extra_defaults={"needs_connection": True, "retry_timer": None, "timeout_timer": None},
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import logging
import threading
import time
from azure.iot.device.iothub.sync_inbox import SyncClientInbox, InboxEmpty

logging.basicConfig(level=logging.DEBUG)


@pytest.mark.describe("SyncClientInbox")
class TestSyncClientInbox(object):
    @pytest.mark.it("Instantiates empty")
    def test_instantiates_empty(self):
        inbox = SyncClientInbox()
        assert inbox.empty()

    @pytest.mark.it("Can be checked regarding whether or not it contains an item")
    def test_check_item_is_in_inbox(self, mocker):
        inbox = SyncClientInbox()
        assert inbox.empty()
        item = mocker.MagicMock()
        assert item not in inbox
        inbox._put(item)
        assert item in inbox

    @pytest.mark.it("Can checked regarding whether or not it is empty")
    def test_check_if_empty(self, mocker):
        inbox = SyncClientInbox()
        assert inbox.empty()
        item = mocker.MagicMock()
        inbox._put(item)
        assert not inbox.empty()
        inbox.get()
        assert inbox.empty()

    @pytest.mark.it("Operates according to FIFO")
    def test_operates_according_to_FIFO(self, mocker):
        inbox = SyncClientInbox()
        item1 = mocker.MagicMock()
        item2 = mocker.MagicMock()
        item3 = mocker.MagicMock()
        inbox._put(item1)
        inbox._put(item2)
        inbox._put(item3)

        assert inbox.get() is item1
        assert inbox.get() is item2
        assert inbox.get() is item3


@pytest.mark.describe("SyncClientInbox - ._put()")
class TestSyncClientInboxPut(object):
    @pytest.mark.it("Adds the given item to the inbox")
    def test_adds_item_to_inbox(self, mocker):
        inbox = SyncClientInbox()
        assert inbox.empty()
        item = mocker.MagicMock()
        inbox._put(item)
        assert not inbox.empty()
        assert item in inbox


@pytest.mark.describe("SyncClientInbox - .get()")
class TestSyncClientInboxGet(object):
    @pytest.mark.it("Returns and removes the next item from the inbox, if there is one")
    def test_removes_item_from_inbox_if_already_there(self, mocker):
        inbox = SyncClientInbox()
        assert inbox.empty()
        item = mocker.MagicMock()
        inbox._put(item)
        assert not inbox.empty()
        retrieved_item = inbox.get()
        assert retrieved_item is item
        assert inbox.empty()

    @pytest.mark.it(
        "Blocks on an empty inbox until an item is available to remove and return, if using blocking mode"
    )
    def test_waits_for_item_to_be_added_if_inbox_empty_in_blocking_mode(self, mocker):
        inbox = SyncClientInbox()
        assert inbox.empty()
        item = mocker.MagicMock()

        def insert_item():
            time.sleep(0.01)  # wait before inserting
            inbox._put(item)

        insertion_thread = threading.Thread(target=insert_item)
        insertion_thread.start()

        retrieved_item = inbox.get(block=True)
        assert retrieved_item is item
        assert inbox.empty()

    @pytest.mark.it(
        "Raises InboxEmpty exception after a timeout while blocking on an empty inbox, if a timeout is specified"
    )
    def test_times_out_while_blocking_if_timeout_specified(self, mocker):
        inbox = SyncClientInbox()
        assert inbox.empty()
        with pytest.raises(InboxEmpty):
            inbox.get(block=True, timeout=0.01)

    @pytest.mark.it(
        "Raises InboxEmpty exception if the inbox is empty, when using non-blocking mode"
    )
    def test_get_raises_empty_if_inbox_empty_in_non_blocking_mode(self):
        inbox = SyncClientInbox()
        assert inbox.empty()
        with pytest.raises(InboxEmpty):
            inbox.get(block=False)


@pytest.mark.describe("SyncClientInbox - .clear()")
class TestSyncClientInboxClear(object):
    @pytest.mark.it("Clears all items from the inbox")
    def test_can_clear_all_items(self, mocker):
        inbox = SyncClientInbox()
        item1 = mocker.MagicMock()
        item2 = mocker.MagicMock()
        item3 = mocker.MagicMock()
        inbox._put(item1)
        inbox._put(item2)
        inbox._put(item3)
        assert not inbox.empty()

        inbox.clear()
        assert inbox.empty()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import six.moves.urllib as urllib
from azure.iot.device.common.pipeline import (
    pipeline_ops_base,
    pipeline_ops_mqtt,
    pipeline_events_mqtt,
    pipeline_thread,
)
from azure.iot.device.common.pipeline.pipeline_stages_base import PipelineStage
from azure.iot.device.provisioning.pipeline import mqtt_topic
from azure.iot.device.provisioning.pipeline import (
    pipeline_events_provisioning,
    pipeline_ops_provisioning,
)
from azure.iot.device import constant as pkg_constant
import json

logger = logging.getLogger(__name__)


class ProvisioningMQTTTranslationStage(PipelineStage):
    """
    PipelineStage which converts other Provisioning pipeline operations into MQTT operations. This stage also
    converts MQTT pipeline events into Provisioning pipeline events.
    """

    def __init__(self):
        super(ProvisioningMQTTTranslationStage, self).__init__()
        self.action_to_topic = {}

    @pipeline_thread.runs_on_pipeline_thread
    def _execute_op(self, op):

        if isinstance(op, pipeline_ops_provisioning.SetProvisioningClientConnectionArgsOperation):
            # get security client args from above, save some, use some to build topic names,
            # always pass it down because MQTT protocol stage will also want to receive these args.

            client_id = op.registration_id
            query_param_seq = [
                ("api-version", pkg_constant.PROVISIONING_API_VERSION),
                ("ClientVersion", pkg_constant.USER_AGENT),
            ]
            username = "{id_scope}/registrations/{registration_id}/{query_params}".format(
                id_scope=op.id_scope,
                registration_id=op.registration_id,
                query_params=urllib.parse.urlencode(query_param_seq),
            )

            hostname = op.provisioning_host

            self.send_worker_op_down(
                worker_op=pipeline_ops_mqtt.SetMQTTConnectionArgsOperation(
                    client_id=client_id,
                    hostname=hostname,
                    username=username,
                    client_cert=op.client_cert,
                    sas_token=op.sas_token,
                    callback=op.callback,
                ),
                op=op,
            )

        elif isinstance(op, pipeline_ops_provisioning.SendRegistrationRequestOperation):
            # Convert Sending the request into MQTT Publish operations
            topic = mqtt_topic.get_topic_for_register(op.request_id)

            # This is an easier way to get the json eventually
            # rather than formatting strings without if else conditions
            registration_payload = DeviceRegistrationPayload(
                registration_id=op.registration_id, custom_payload=op.request_payload
            )

            self.send_worker_op_down(
                worker_op=pipeline_ops_mqtt.MQTTPublishOperation(
                    topic=topic,
                    payload=registration_payload.get_json_string(),
                    callback=op.callback,
                ),
                op=op,
            )

        elif isinstance(op, pipeline_ops_provisioning.SendQueryRequestOperation):
            # Convert Sending the request into MQTT Publish operations
            topic = mqtt_topic.get_topic_for_query(op.request_id, op.operation_id)
            self.send_worker_op_down(
                worker_op=pipeline_ops_mqtt.MQTTPublishOperation(
                    topic=topic, payload=op.request_payload, callback=op.callback
                ),
                op=op,
            )

        elif isinstance(op, pipeline_ops_base.EnableFeatureOperation):
            # Enabling for register gets translated into an MQTT subscribe operation
            topic = mqtt_topic.get_topic_for_subscribe()
            self.send_worker_op_down(
                worker_op=pipeline_ops_mqtt.MQTTSubscribeOperation(
                    topic=topic, callback=op.callback
                ),
                op=op,
            )

        elif isinstance(op, pipeline_ops_base.DisableFeatureOperation):
            # Disabling a register response gets turned into an MQTT unsubscribe operation
            topic = mqtt_topic.get_topic_for_subscribe()
            self.send_worker_op_down(
                worker_op=pipeline_ops_mqtt.MQTTUnsubscribeOperation(
                    topic=topic, callback=op.callback
                ),
                op=op,
            )

        else:
            # All other operations get passed down
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):
        """
        Pipeline Event handler function to convert incoming MQTT messages into the appropriate DPS
        events, based on the topic of the message
        """
        if isinstance(event, pipeline_events_mqtt.IncomingMQTTMessageEvent):
            topic = event.topic

            if mqtt_topic.is_dps_response_topic(topic):
                logger.info(
                    "Received payload:{payload} on topic:{topic}".format(
                        payload=event.payload, topic=topic
                    )
                )
                key_values = mqtt_topic.extract_properties_from_topic(topic)
                status_code = mqtt_topic.extract_status_code_from_topic(topic)
                request_id = key_values["rid"][0]
                if event.payload is not None:
                    response = event.payload.decode("utf-8")
                # Extract pertinent information from mqtt topic
                # like status code request_id and send it upwards.
                self.send_event_up(
                    pipeline_events_provisioning.RegistrationResponseEvent(
                        request_id, status_code, key_values, response
                    )
                )
            else:
                logger.warning("Unknown topic: {} passing up to next handler".format(topic))
                self.send_event_up(event)

        else:
            # all other messages get passed up
            self.send_event_up(event)


class DeviceRegistrationPayload(object):
    """
    The class representing the payload that needs to be sent to the service.
    """

    def __init__(self, registration_id, custom_payload=None):
        # This is not a convention to name variables in python but the
        # DPS service spec needs the name to be exact for it to work
        self.registrationId = registration_id
        self.payload = custom_payload

    def get_json_string(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest


class SharedAuthenticationProviderInstantiationTests(object):
    @pytest.mark.it("Sets the hostname parameter as an instance attribute")
    def test_hostname(self, auth_provider, hostname):
        assert auth_provider.hostname == hostname

    @pytest.mark.it("Sets the device_id parameter as an instance attribute")
    def test_device_id(self, auth_provider, device_id):
        assert auth_provider.device_id == device_id

    @pytest.mark.it("Sets the module_id parameter as an instance attribute")
    def test_module_id(self, auth_provider, module_id):
        assert auth_provider.module_id == module_id


class SharedBaseRenewableAuthenticationProviderInstantiationTests(
    SharedAuthenticationProviderInstantiationTests
):
    # TODO: Complete this testclass after refactoring the class under test
    pass



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import sys
import pytest
import logging
from azure.iot.device.common.pipeline import pipeline_events_base
from tests.common.pipeline import pipeline_data_object_test

logging.basicConfig(level=logging.DEBUG)
this_module = sys.modules[__name__]


@pytest.mark.describe("PipelineEvent")
class TestPipelineOperation(object):
    @pytest.mark.it("Can't be instantiated")
    def test_instantiate(self):
        with pytest.raises(TypeError):
            pipeline_events_base.PipelineEvent()


pipeline_data_object_test.add_event_test(
    cls=pipeline_events_base.ResponseEvent,
    module=this_module,
    positional_arguments=["request_id", "status_code", "response_body"],
    keyword_arguments={},
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import pytest
import threading
from tests.common.pipeline import helpers
from azure.iot.device.common import handle_exceptions
from azure.iot.device.common.pipeline import (
    pipeline_events_base,
    pipeline_ops_base,
    pipeline_thread,
)


class FakeEvent(pipeline_events_base.PipelineEvent):
    def __init__(self):
        super(FakeEvent, self).__init__()


@pytest.fixture
def arbitrary_event():
    return FakeEvent()


class FakeOperation(pipeline_ops_base.PipelineOperation):
    def __init__(self, callback=None):
        super(FakeOperation, self).__init__(callback=callback)


@pytest.fixture
def arbitrary_op(mocker):
    op = FakeOperation(callback=mocker.MagicMock())
    op.name = "arbitrary_op"
    return op


@pytest.fixture
def fake_pipeline_thread():
    """
    This fixture mocks out the thread name so that the pipeline decorators
    use to assert that you are in a pipeline thread.
    """
    this_thread = threading.current_thread()
    old_name = this_thread.name

    this_thread.name = "pipeline"
    yield
    this_thread.name = old_name


@pytest.fixture
def fake_non_pipeline_thread():
    """
    This fixture sets thread name to something other than "pipeline" to force asserts
    """
    this_thread = threading.current_thread()
    old_name = this_thread.name

    this_thread.name = "not pipeline"
    yield
    this_thread.name = old_name


@pytest.fixture
def unhandled_error_handler(mocker):
    return mocker.patch.object(handle_exceptions, "handle_background_exception")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest


@pytest.fixture
def hostname():
    return "__FAKE_HOSTNAME__"


@pytest.fixture
def device_id():
    return "__FAKE_DEVICE_ID__"


@pytest.fixture
def module_id():
    return "__FAKE_MODULE__ID__"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import logging
import pytest
import sys
import six
from azure.iot.device.common import transport_exceptions, handle_exceptions
from azure.iot.device.common.pipeline import (
    pipeline_ops_base,
    pipeline_stages_base,
    pipeline_ops_mqtt,
    pipeline_events_mqtt,
    pipeline_stages_mqtt,
    pipeline_exceptions,
    config,
)
from tests.common.pipeline.helpers import (
    assert_callback_failed,
    assert_callback_succeeded,
    all_common_ops,
    all_common_events,
    all_except,
    StageTestBase,
)
from tests.common.pipeline import pipeline_stage_test

logging.basicConfig(level=logging.DEBUG)


# Normally, all arbitrary exceptions, representing some kind nonspecific unexpected and unhandled
# exception that are used in tests should come from the top level fixture "arbitrary_exception",
# however, since fixtures cannot be used in parametrization of tests, we need to define a custom
# Exception class here.
class SomeException(Exception):
    pass


# This fixture makes it look like all test in this file  tests are running
# inside the pipeline thread.  Because this is an autouse fixture, we
# manually add it to the individual test.py files that need it.  If,
# instead, we had added it to some conftest.py, it would be applied to
# every tests in every file and we don't want that.
@pytest.fixture(autouse=True)
def apply_fake_pipeline_thread(fake_pipeline_thread):
    pass


class FakeOperation(pipeline_ops_base.PipelineOperation):
    pass


this_module = sys.modules[__name__]

fake_client_id = "__fake_client_id__"
fake_hostname = "__fake_hostname__"
fake_username = "__fake_username__"
fake_ca_cert = "__fake_ca_cert__"
fake_sas_token = "__fake_sas_token__"
fake_topic = "__fake_topic__"
fake_payload = "__fake_payload__"
fake_certificate = "__fake_certificate__"
fake_boolean = "__fake_boolean__"

ops_handled_by_this_stage = [
    pipeline_ops_base.ConnectOperation,
    pipeline_ops_base.DisconnectOperation,
    pipeline_ops_base.ReconnectOperation,
    pipeline_ops_base.UpdateSasTokenOperation,
    pipeline_ops_mqtt.SetMQTTConnectionArgsOperation,
    pipeline_ops_mqtt.MQTTPublishOperation,
    pipeline_ops_mqtt.MQTTSubscribeOperation,
    pipeline_ops_mqtt.MQTTUnsubscribeOperation,
]

events_handled_by_this_stage = []

# TODO: Potentially refactor this out to package level class that can be inherited
pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_mqtt.MQTTTransportStage,
    module=this_module,
    all_ops=all_common_ops,
    handled_ops=ops_handled_by_this_stage,
    all_events=all_common_events,
    handled_events=events_handled_by_this_stage,
    methods_that_enter_pipeline_thread=[
        "_on_mqtt_message_received",
        "_on_mqtt_connected",
        "_on_mqtt_connection_failure",
        "_on_mqtt_disconnected",
    ],
)


@pytest.fixture
def transport(mocker):
    mocker.patch(
        "azure.iot.device.common.pipeline.pipeline_stages_mqtt.MQTTTransport", autospec=True
    )
    return pipeline_stages_mqtt.MQTTTransport


@pytest.fixture
def op_set_connection_args(mocker):
    return pipeline_ops_mqtt.SetMQTTConnectionArgsOperation(
        client_id=fake_client_id,
        hostname=fake_hostname,
        username=fake_username,
        ca_cert=fake_ca_cert,
        client_cert=fake_certificate,
        sas_token=fake_sas_token,
        callback=mocker.MagicMock(),
    )


@pytest.fixture
def op_connect(mocker):
    return pipeline_ops_base.ConnectOperation(callback=mocker.MagicMock())


@pytest.fixture
def op_reconnect(mocker):
    return pipeline_ops_base.ReconnectOperation(callback=mocker.MagicMock())


@pytest.fixture
def op_disconnect(mocker):
    return pipeline_ops_base.DisconnectOperation(callback=mocker.MagicMock())


@pytest.fixture
def op_publish(mocker):
    return pipeline_ops_mqtt.MQTTPublishOperation(
        topic=fake_topic, payload=fake_payload, callback=mocker.MagicMock()
    )


@pytest.fixture
def op_subscribe(mocker):
    return pipeline_ops_mqtt.MQTTSubscribeOperation(topic=fake_topic, callback=mocker.MagicMock())


@pytest.fixture
def op_unsubscribe(mocker):
    return pipeline_ops_mqtt.MQTTUnsubscribeOperation(topic=fake_topic, callback=mocker.MagicMock())


@pytest.fixture
def create_transport(stage, transport, op_set_connection_args):
    stage.run_op(op_set_connection_args)


# TODO: This should be a package level class inherited by all .run_op() tests in all stages
class RunOpTests(object):
    @pytest.mark.it(
        "Completes the operation with failure if an unexpected Exception is raised while executing the operation"
    )
    def test_completes_operation_with_error(self, mocker, stage, arbitrary_exception, arbitrary_op):
        stage._execute_op = mocker.MagicMock(side_effect=arbitrary_exception)

        stage.run_op(arbitrary_op)
        assert arbitrary_op.callback.call_count == 1
        assert arbitrary_op.callback.call_args == mocker.call(
            arbitrary_op, error=arbitrary_exception
        )

    @pytest.mark.it(
        "Allows any BaseException that was raised during execution of the operation to propogate"
    )
    def test_base_exception_propogates(self, mocker, stage, arbitrary_base_exception):
        execution_exception = arbitrary_base_exception
        mock_op = mocker.MagicMock()
        stage._execute_op = mocker.MagicMock(side_effect=execution_exception)

        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.run_op(mock_op)
        assert e_info.value is arbitrary_base_exception


class MQTTTransportStageTestBase(StageTestBase):
    """
    Base class for all tests in this file.
    """

    @pytest.fixture
    def stage(self):
        return pipeline_stages_mqtt.MQTTTransportStage()

    @pytest.fixture(autouse=True)
    def fix_base_fixtures_for_these_tests(self, stage, stage_base_configuration, mocker):
        # run this fixture after stage_base_configuration to further configure the pipeline for these tests

        stage.next = None

        mocker.spy(stage.pipeline_root, "handle_pipeline_event")
        mocker.spy(stage.pipeline_root, "on_connected")
        mocker.spy(stage.pipeline_root, "on_disconnected")

        mocker.spy(stage, "_on_mqtt_connected")
        mocker.spy(stage, "_on_mqtt_connection_failure")
        mocker.spy(stage, "_on_mqtt_disconnected")


@pytest.mark.describe(
    "MQTTTransportStage - .run_op() -- called with pipeline_ops_mqtt.SetMQTTConnectionArgsOperation"
)
class TestMQTTTransportStageRunOpWithSetConnectionArgs(MQTTTransportStageTestBase, RunOpTests):
    @pytest.mark.it("Creates an MQTTTransport object")
    def test_creates_transport(self, stage, transport, op_set_connection_args):
        stage.run_op(op_set_connection_args)
        assert transport.call_count == 1

    @pytest.mark.it(
        "Initializes the MQTTTransport object with the passed client_id, hostname, username, ca_cert and x509_cert"
    )
    def test_passes_right_params(self, stage, transport, mocker, op_set_connection_args):
        stage.run_op(op_set_connection_args)
        assert transport.call_args == mocker.call(
            client_id=fake_client_id,
            hostname=fake_hostname,
            username=fake_username,
            ca_cert=fake_ca_cert,
            x509_cert=fake_certificate,
            websockets=False,
        )

    @pytest.mark.it(
        "Initializes the MQTTTransport object with the passed websockets from setting the PipelineRootStage config"
    )
    def test_receives_correct_config(self, stage, transport, mocker, op_set_connection_args):
        stage.pipeline_root = pipeline_stages_base.PipelineRootStage(
            config.BasePipelineConfig(websockets="__fake_boolean__")
        )
        stage.run_op(op_set_connection_args)
        assert transport.call_args == mocker.call(
            client_id=fake_client_id,
            hostname=fake_hostname,
            username=fake_username,
            ca_cert=fake_ca_cert,
            x509_cert=fake_certificate,
            websockets="__fake_boolean__",
        )

    @pytest.mark.it("Sets handlers on the transport")
    def test_sets_parameters(self, stage, transport, mocker, op_set_connection_args):
        stage.run_op(op_set_connection_args)
        assert transport.return_value.on_mqtt_disconnected_handler == stage._on_mqtt_disconnected
        assert transport.return_value.on_mqtt_connected_handler == stage._on_mqtt_connected
        assert (
            transport.return_value.on_mqtt_connection_failure_handler
            == stage._on_mqtt_connection_failure
        )
        assert (
            transport.return_value.on_mqtt_message_received_handler
            == stage._on_mqtt_message_received
        )

    @pytest.mark.it("Sets the pending connection op tracker to None")
    def test_pending_conn_op(self, stage, transport, op_set_connection_args):
        stage.run_op(op_set_connection_args)
        assert stage._pending_connection_op is None

    @pytest.mark.it("Completes the operation with success, upon successful execution")
    def test_succeeds(self, stage, transport, op_set_connection_args):
        stage.run_op(op_set_connection_args)
        assert_callback_succeeded(op=op_set_connection_args)


@pytest.mark.describe("MQTTTransportStage - .run_op() -- called with ConnectOperation")
class TestMQTTTransportStageExecuteOpWithConnect(MQTTTransportStageTestBase, RunOpTests):
    @pytest.mark.it("Sets the ConnectOperation as the pending connection operation")
    def test_sets_pending_operation(self, stage, create_transport, op_connect):
        stage.run_op(op_connect)
        assert stage._pending_connection_op is op_connect

    @pytest.mark.it("Cancels any already pending connection operation")
    @pytest.mark.parametrize(
        "pending_connection_op",
        [
            pytest.param(pipeline_ops_base.ConnectOperation(1), id="Pending ConnectOperation"),
            pytest.param(pipeline_ops_base.ReconnectOperation(1), id="Pending ReconnectOperation"),
            pytest.param(
                pipeline_ops_base.DisconnectOperation(1), id="Pending DisconnectOperation"
            ),
        ],
    )
    def test_pending_operation_cancelled(
        self, mocker, stage, create_transport, op_connect, pending_connection_op
    ):
        pending_connection_op.callback = mocker.MagicMock()
        stage._pending_connection_op = pending_connection_op
        stage.run_op(op_connect)

        # Callback has been completed, with an OpeationCancelled exception set indicating early cancellation
        assert_callback_failed(
            op=pending_connection_op, error=pipeline_exceptions.OperationCancelled
        )

        # New operation is now the pending operation
        assert stage._pending_connection_op is op_connect

    @pytest.mark.it("Does an MQTT connect via the MQTTTransport")
    def test_mqtt_connect(self, mocker, stage, create_transport, op_connect):
        stage.run_op(op_connect)
        assert stage.transport.connect.call_count == 1
        assert stage.transport.connect.call_args == mocker.call(password=stage.sas_token)

    @pytest.mark.it(
        "Fails the operation and resets the pending connection operation to None, if there is a failure connecting in the MQTTTransport"
    )
    def test_fails_operation(self, stage, create_transport, op_connect, arbitrary_exception):
        stage.transport.connect.side_effect = arbitrary_exception
        stage.run_op(op_connect)
        assert_callback_failed(op=op_connect, error=arbitrary_exception)
        assert stage._pending_connection_op is None


@pytest.mark.describe("MQTTTransportStage - .run_op() -- called with ReconnectOperation")
class TestMQTTTransportStageExecuteOpWithReconnect(MQTTTransportStageTestBase, RunOpTests):
    @pytest.mark.it("Sets the ReconnectOperation as the pending connection operation")
    def test_sets_pending_operation(self, stage, create_transport, op_reconnect):
        stage.run_op(op_reconnect)
        assert stage._pending_connection_op is op_reconnect

    @pytest.mark.it("Cancels any already pending connection operation")
    @pytest.mark.parametrize(
        "pending_connection_op",
        [
            pytest.param(pipeline_ops_base.ConnectOperation(1), id="Pending ConnectOperation"),
            pytest.param(pipeline_ops_base.ReconnectOperation(1), id="Pending ReconnectOperation"),
            pytest.param(
                pipeline_ops_base.DisconnectOperation(1), id="Pending DisconnectOperation"
            ),
        ],
    )
    def test_pending_operation_cancelled(
        self, mocker, stage, create_transport, op_reconnect, pending_connection_op
    ):
        pending_connection_op.callback = mocker.MagicMock()
        stage._pending_connection_op = pending_connection_op
        stage.run_op(op_reconnect)

        # Callback has been completed, with an OperationCancelled exception set indicating early cancellation
        assert_callback_failed(
            op=pending_connection_op, error=pipeline_exceptions.OperationCancelled
        )

        # New operation is now the pending operation
        assert stage._pending_connection_op is op_reconnect

    @pytest.mark.it("Does an MQTT reconnect via the MQTTTransport")
    def test_mqtt_reconnect(self, mocker, stage, create_transport, op_reconnect):
        stage.run_op(op_reconnect)
        assert stage.transport.reconnect.call_count == 1
        assert stage.transport.reconnect.call_args == mocker.call(password=stage.sas_token)

    @pytest.mark.it(
        "Fails the operation and resets the pending connection operation to None, if there is a failure reconnecting in the MQTTTransport"
    )
    def test_fails_operation(
        self, mocker, stage, create_transport, op_reconnect, arbitrary_exception
    ):
        stage.transport.reconnect.side_effect = arbitrary_exception
        stage.run_op(op_reconnect)
        assert_callback_failed(op=op_reconnect, error=arbitrary_exception)
        assert stage._pending_connection_op is None


@pytest.mark.describe("MQTTTransportStage - .run_op() -- called with DisconnectOperation")
class TestMQTTTransportStageExecuteOpWithDisconnect(MQTTTransportStageTestBase, RunOpTests):
    @pytest.mark.it("Sets the DisconnectOperation as the pending connection operation")
    def test_sets_pending_operation(self, stage, create_transport, op_disconnect):
        stage.run_op(op_disconnect)
        assert stage._pending_connection_op is op_disconnect

    @pytest.mark.it("Cancels any already pending connection operation")
    @pytest.mark.parametrize(
        "pending_connection_op",
        [
            pytest.param(pipeline_ops_base.ConnectOperation(1), id="Pending ConnectOperation"),
            pytest.param(pipeline_ops_base.ReconnectOperation(1), id="Pending ReconnectOperation"),
            pytest.param(
                pipeline_ops_base.DisconnectOperation(1), id="Pending DisconnectOperation"
            ),
        ],
    )
    def test_pending_operation_cancelled(
        self, mocker, stage, create_transport, op_disconnect, pending_connection_op
    ):
        pending_connection_op.callback = mocker.MagicMock()
        stage._pending_connection_op = pending_connection_op
        stage.run_op(op_disconnect)

        # Callback has been completed, with an OperationCancelled exception set indicating early cancellation
        assert_callback_failed(
            op=pending_connection_op, error=pipeline_exceptions.OperationCancelled
        )

        # New operation is now the pending operation
        assert stage._pending_connection_op is op_disconnect

    @pytest.mark.it("Does an MQTT disconnect via the MQTTTransport")
    def test_mqtt_disconnect(self, mocker, stage, create_transport, op_disconnect):
        stage.run_op(op_disconnect)
        assert stage.transport.disconnect.call_count == 1
        assert stage.transport.disconnect.call_args == mocker.call()

    @pytest.mark.it(
        "Fails the operation and resets the pending connection operation to None, if there is a failure disconnecting in the MQTTTransport"
    )
    def test_fails_operation(
        self, mocker, stage, create_transport, op_disconnect, arbitrary_exception
    ):
        stage.transport.disconnect.side_effect = arbitrary_exception
        stage.run_op(op_disconnect)
        assert_callback_failed(op=op_disconnect, error=arbitrary_exception)
        assert stage._pending_connection_op is None


@pytest.mark.describe("MQTTTransportStage - .run_op() -- called with MQTTPublishOperation")
class TestMQTTTransportStageExecuteOpWithMQTTPublishOperation(
    MQTTTransportStageTestBase, RunOpTests
):
    @pytest.mark.it("Does an MQTT publish via the MQTTTransport")
    def test_mqtt_publish(self, mocker, stage, create_transport, op_publish):
        stage.run_op(op_publish)
        assert stage.transport.publish.call_count == 1
        assert stage.transport.publish.call_args == mocker.call(
            topic=op_publish.topic, payload=op_publish.payload, callback=mocker.ANY
        )

    @pytest.mark.it(
        "Completes the operation with success, upon successful completion of the MQTT publish"
    )
    def test_complete(self, mocker, stage, create_transport, op_publish):
        # Begin publish
        stage.run_op(op_publish)

        # Trigger publish completion
        stage.transport.publish.call_args[1]["callback"]()

        assert_callback_succeeded(op=op_publish)


@pytest.mark.describe("MQTTTransportStage - .run_op() -- called with MQTTSubscribeOperation")
class TestMQTTTransportStageExecuteOpWithMQTTSubscribeOperation(
    MQTTTransportStageTestBase, RunOpTests
):
    @pytest.mark.it("Does an MQTT subscribe via the MQTTTransport")
    def test_mqtt_publish(self, mocker, stage, create_transport, op_subscribe):
        stage.run_op(op_subscribe)
        assert stage.transport.subscribe.call_count == 1
        assert stage.transport.subscribe.call_args == mocker.call(
            topic=op_subscribe.topic, callback=mocker.ANY
        )

    @pytest.mark.it(
        "Completes the operation with success, upon successful completion of the MQTT subscribe"
    )
    def test_complete(self, mocker, stage, create_transport, op_subscribe):
        # Begin subscribe
        stage.run_op(op_subscribe)

        # Trigger subscribe completion
        stage.transport.subscribe.call_args[1]["callback"]()

        assert_callback_succeeded(op=op_subscribe)


@pytest.mark.describe("MQTTTransportStage - .run_op() -- called with MQTTUnsubscribeOperation")
class TestMQTTTransportStageExecuteOpWithMQTTUnsubscribeOperation(
    MQTTTransportStageTestBase, RunOpTests
):
    @pytest.mark.it("Does an MQTT unsubscribe via the MQTTTransport")
    def test_mqtt_publish(self, mocker, stage, create_transport, op_unsubscribe):
        stage.run_op(op_unsubscribe)
        assert stage.transport.unsubscribe.call_count == 1
        assert stage.transport.unsubscribe.call_args == mocker.call(
            topic=op_unsubscribe.topic, callback=mocker.ANY
        )

    @pytest.mark.it(
        "Completes the operation with success, upon successful completion of the MQTT unsubscribe"
    )
    def test_complete(self, mocker, stage, create_transport, op_unsubscribe):
        # Begin unsubscribe
        stage.run_op(op_unsubscribe)

        # Trigger unsubscribe completion
        stage.transport.unsubscribe.call_args[1]["callback"]()

        assert_callback_succeeded(op=op_unsubscribe)


@pytest.mark.describe("MQTTTransportStage - .run_op() -- called with UpdateSasTokenOperation")
class TestMQTTTransportStageExecuteOpWithUpdateSasTokenoperation(
    MQTTTransportStageTestBase, RunOpTests
):
    @pytest.mark.it("Saves the token and completes immediately")
    def test_mqtt_publish(self, mocker, stage, create_transport):
        cb = mocker.MagicMock()
        op_update_sas_token = pipeline_ops_base.UpdateSasTokenOperation(
            sas_token=fake_sas_token, callback=cb
        )
        stage.run_op(op_update_sas_token)
        assert_callback_succeeded(op_update_sas_token)
        assert stage.sas_token == fake_sas_token


@pytest.mark.describe("MQTTTransportStage - EVENT: MQTT message received")
class TestMQTTTransportStageProtocolClientEvents(MQTTTransportStageTestBase):
    @pytest.mark.it("Fires an IncomingMQTTMessageEvent event for each MQTT message received")
    def test_incoming_message_handler(self, stage, create_transport, mocker):
        stage.transport.on_mqtt_message_received_handler(topic=fake_topic, payload=fake_payload)
        assert stage.previous.handle_pipeline_event.call_count == 1
        call_arg = stage.previous.handle_pipeline_event.call_args[0][0]
        assert isinstance(call_arg, pipeline_events_mqtt.IncomingMQTTMessageEvent)

    @pytest.mark.it("Passes topic and payload as part of the IncomingMQTTMessageEvent event")
    def test_verify_incoming_message_attributes(self, stage, create_transport, mocker):
        stage.transport.on_mqtt_message_received_handler(topic=fake_topic, payload=fake_payload)
        call_arg = stage.previous.handle_pipeline_event.call_args[0][0]
        assert call_arg.payload == fake_payload
        assert call_arg.topic == fake_topic


@pytest.mark.describe("MQTTTransportStage - EVENT: MQTT connected")
class TestMQTTTransportStageOnConnected(MQTTTransportStageTestBase):
    @pytest.mark.it("Calls self.on_connected when the transport connected event fires")
    @pytest.mark.parametrize(
        "pending_connection_op",
        [
            pytest.param(None, id="No pending operation"),
            pytest.param(pipeline_ops_base.ConnectOperation(1), id="Pending ConnectOperation"),
            pytest.param(pipeline_ops_base.ReconnectOperation(1), id="Pending ReconnectOperation"),
            pytest.param(
                pipeline_ops_base.DisconnectOperation(1), id="Pending DisconnectOperation"
            ),
        ],
    )
    def test_connected_handler(self, stage, create_transport, pending_connection_op):
        stage._pending_connection_op = pending_connection_op
        assert stage.previous.on_connected.call_count == 0
        stage.transport.on_mqtt_connected_handler()
        assert stage.previous.on_connected.call_count == 1

    @pytest.mark.it(
        "Completes a pending ConnectOperation with success when the transport connected event fires"
    )
    def test_completes_pending_connect_op(self, mocker, stage, create_transport):
        op = pipeline_ops_base.ConnectOperation(callback=mocker.MagicMock())
        stage.run_op(op)
        assert op.callback.call_count == 0
        assert stage._pending_connection_op is op
        stage.transport.on_mqtt_connected_handler()
        assert_callback_succeeded(op=op)
        assert stage._pending_connection_op is None

    @pytest.mark.it(
        "Completes a pending ReconnectOperation with success when the transport connected event fires"
    )
    def test_completes_pending_reconnect_op(self, mocker, stage, create_transport):
        op = pipeline_ops_base.ReconnectOperation(callback=mocker.MagicMock())
        stage.run_op(op)
        assert op.callback.call_count == 0
        assert stage._pending_connection_op is op
        stage.transport.on_mqtt_connected_handler()
        assert_callback_succeeded(op=op)
        assert stage._pending_connection_op is None

    @pytest.mark.it(
        "Ignores a pending DisconnectOperation when the transport connected event fires"
    )
    def test_ignores_pending_disconnect_op(self, mocker, stage, create_transport):
        op = pipeline_ops_base.DisconnectOperation(callback=mocker.MagicMock())
        stage.run_op(op)
        assert op.callback.call_count == 0
        assert stage._pending_connection_op is op
        stage.transport.on_mqtt_connected_handler()
        # handler did NOT trigger a callback
        assert op.callback.call_count == 0
        assert stage._pending_connection_op is op


@pytest.mark.describe("MQTTTarnsportStage - EVENT: MQTT connection failure")
class TestMQTTTransportStageOnConnectionFailure(MQTTTransportStageTestBase):
    @pytest.mark.it(
        "Does not call self.on_connected when the transport connection failure event fires"
    )
    @pytest.mark.parametrize(
        "pending_connection_op",
        [
            pytest.param(None, id="No pending operation"),
            pytest.param(pipeline_ops_base.ConnectOperation(1), id="Pending ConnectOperation"),
            pytest.param(pipeline_ops_base.ReconnectOperation(1), id="Pending ReconnectOperation"),
            pytest.param(
                pipeline_ops_base.DisconnectOperation(1), id="Pending DisconnectOperation"
            ),
        ],
    )
    def test_does_not_call_connected_handler(
        self, stage, create_transport, arbitrary_exception, pending_connection_op
    ):
        # This test is testing negative space - something the function does NOT do - rather than something it does
        stage._pending_connection_op = pending_connection_op
        assert stage.previous.on_connected.call_count == 0
        stage.transport.on_mqtt_connection_failure_handler(arbitrary_exception)
        assert stage.previous.on_connected.call_count == 0

    @pytest.mark.it("Fails a pending ConnectOperation if the connection failure event fires")
    def test_fails_pending_connect_op(self, mocker, stage, create_transport, arbitrary_exception):
        op = pipeline_ops_base.ConnectOperation(callback=mocker.MagicMock())
        stage.run_op(op)
        assert op.callback.call_count == 0
        assert stage._pending_connection_op is op
        stage.transport.on_mqtt_connection_failure_handler(arbitrary_exception)
        assert_callback_failed(op=op, error=arbitrary_exception)
        assert stage._pending_connection_op is None

    @pytest.mark.it("Fails a pending ReconnectOperation if the connection failure event fires")
    def test_fails_pending_reconnect_op(self, mocker, stage, create_transport, arbitrary_exception):
        op = pipeline_ops_base.ReconnectOperation(callback=mocker.MagicMock())
        stage.run_op(op)
        assert op.callback.call_count == 0
        assert stage._pending_connection_op is op
        stage.transport.on_mqtt_connection_failure_handler(arbitrary_exception)
        assert_callback_failed(op=op, error=arbitrary_exception)
        assert stage._pending_connection_op is None

    @pytest.mark.it("Ignores a pending DisconnectOperation if the connection failure event fires")
    def test_ignores_pending_disconnect_op(
        self, mocker, stage, create_transport, arbitrary_exception
    ):
        op = pipeline_ops_base.DisconnectOperation(callback=mocker.MagicMock())
        stage.run_op(op)
        assert op.callback.call_count == 0
        assert stage._pending_connection_op is op
        stage.transport.on_mqtt_connection_failure_handler(arbitrary_exception)
        # Assert nothing changed about the operation
        assert op.callback.call_count == 0
        assert stage._pending_connection_op is op

    @pytest.mark.it(
        "Triggers the unhandled exception handler (with error cause) when the connection failure is unexpected"
    )
    @pytest.mark.parametrize(
        "pending_connection_op",
        [
            pytest.param(None, id="No pending operation"),
            pytest.param(
                pipeline_ops_base.DisconnectOperation(1), id="Pending DisconnectOperation"
            ),
        ],
    )
    def test_unexpected_connection_failure(
        self, mocker, stage, create_transport, arbitrary_exception, pending_connection_op
    ):
        # A connection failure is unexpected if there is not a pending Connect/Reconnect operation
        # i.e. "Why did we get a connection failure? We weren't even trying to connect!"
        mock_handler = mocker.patch.object(handle_exceptions, "handle_background_exception")
        stage._pending_connection_operation = pending_connection_op
        stage.transport.on_mqtt_connection_failure_handler(arbitrary_exception)
        assert mock_handler.call_count == 1
        assert mock_handler.call_args[0][0] is arbitrary_exception


@pytest.mark.describe("MQTTTransportStage - EVENT: MQTT disconnected")
class TestMQTTTransportStageOnDisconnected(MQTTTransportStageTestBase):
    @pytest.mark.it("Calls self.on_disconnected when the transport disconnected event fires")
    @pytest.mark.parametrize(
        "cause",
        [
            pytest.param(None, id="No error cause"),
            pytest.param(SomeException(), id="With error cause"),
        ],
    )
    @pytest.mark.parametrize(
        "pending_connection_op",
        [
            pytest.param(None, id="No pending operation"),
            pytest.param(pipeline_ops_base.ConnectOperation(1), id="Pending ConnectOperation"),
            pytest.param(pipeline_ops_base.ReconnectOperation(1), id="Pending ReconnectOperation"),
            pytest.param(
                pipeline_ops_base.DisconnectOperation(1), id="Pending DisconnectOperation"
            ),
        ],
    )
    def test_disconnected_handler(self, stage, create_transport, pending_connection_op, cause):
        stage._pending_connection_op = pending_connection_op
        assert stage.previous.on_disconnected.call_count == 0
        stage.transport.on_mqtt_disconnected_handler(cause)
        assert stage.previous.on_disconnected.call_count == 1

    @pytest.mark.it(
        "Completes a pending DisconnectOperation with success when the transport disconnected event fires without an error cause"
    )
    def test_compltetes_pending_disconnect_op_when_no_error(self, mocker, stage, create_transport):
        op = pipeline_ops_base.DisconnectOperation(callback=mocker.MagicMock())
        stage.run_op(op)
        assert op.callback.call_count == 0
        assert stage._pending_connection_op is op
        stage.transport.on_mqtt_disconnected_handler(None)
        assert_callback_succeeded(op=op)
        assert stage._pending_connection_op is None

    @pytest.mark.it(
        "Completes a pending DisconnectOperation with success when the transport disconnected event fires with an error cause"
    )
    def test_completes_pending_disconnect_op_with_error(
        self, mocker, stage, create_transport, arbitrary_exception
    ):
        op = pipeline_ops_base.DisconnectOperation(callback=mocker.MagicMock())
        stage.run_op(op)
        assert op.callback.call_count == 0
        assert stage._pending_connection_op is op
        stage.transport.on_mqtt_disconnected_handler(arbitrary_exception)
        assert_callback_succeeded(op=op)
        assert stage._pending_connection_op is None

    @pytest.mark.it(
        "Completes an unrelated pending operation when the transport disconnected event fires"
    )
    @pytest.mark.parametrize(
        "cause",
        [
            pytest.param(None, id="No error cause"),
            pytest.param(SomeException(), id="With error cause"),
        ],
    )
    @pytest.mark.parametrize(
        "pending_connection_op",
        [
            pytest.param(pipeline_ops_base.ConnectOperation(None), id="Pending ConnectOperation"),
            pytest.param(
                pipeline_ops_base.ReconnectOperation(None), id="Pending ReconnectOperation"
            ),
        ],
    )
    def test_completes_unrelated_op(
        self, mocker, stage, create_transport, pending_connection_op, cause
    ):
        pending_connection_op.callback = mocker.MagicMock()
        pending_connection_op.completed = False
        stage._pending_connection_op = pending_connection_op
        stage.transport.on_mqtt_disconnected_handler(cause)
        assert stage._pending_connection_op is None
        assert pending_connection_op.callback.call_count == 1
        if cause:
            assert pending_connection_op.callback.call_args[1]["error"] == cause
        else:
            assert (
                type(pending_connection_op.callback.call_args[1]["error"])
                == transport_exceptions.ConnectionDroppedError
            )

    @pytest.mark.it(
        "Triggers the unhandled exception handler (with ConnectionDroppedError) when the disconnect is unexpected"
    )
    @pytest.mark.parametrize(
        "cause",
        [
            pytest.param(None, id="No error cause"),
            pytest.param(SomeException(), id="With error cause"),
        ],
    )
    def test_unexpected_disconnect(self, mocker, stage, create_transport, cause):
        # A disconnect is unexpected when there is no pending operation, or a pending, non-Disconnect operation
        mock_handler = mocker.patch.object(handle_exceptions, "handle_background_exception")
        stage.transport.on_mqtt_disconnected_handler(cause)
        assert mock_handler.call_count == 1
        assert isinstance(mock_handler.call_args[0][0], transport_exceptions.ConnectionDroppedError)
        assert mock_handler.call_args[0][0].__cause__ is cause
